"""

Constants necessary for EMirror, seperated into classes

"""

from re import compile

# Program Information (PI*)

PIVersion = "2.1.21"
PIHeader  = "\nECLiPt-Mirror Version " + PIVersion + " (c) 1998-2002 Martin Preishuber\n"
PIUsage   = "Usage: emirror [options] URL\n\n" + \
          "Mandatory arguments to long options are mandatory for short options too.\n"
PIOptions = "Startup:\n" + \
          "  -V, --version   display the version of ECLiPt Mirror and exit.\n" + \
          "  -h, --help      print this help.\n" + \
          "  -v, --verbose   be verbose.\n\n" + \
          "Logging, input and output files:\n" + \
          "  -o, --output-file=FILE             log messages to FILE.\n" + \
          "  -O, --output-directory=DIRECTORY   put logfiles into DIRECTORY.\n" + \
          "  -f, --output-format=FORMAT         set logfile format to FORMAT.\n" + \
          "  -c, --config-file=FILE             load configuration information from FILE.\n" + \
          "  -d, --config-directory=DIRECTORY   load configuration files from DIRECTORY.\n" + \
          "  -n, --index-directory=DIRECTORY    set index-file/log-file directory to DIR.\n" + \
          "      --category=TEXT                set category of the mirror to TEXT.\n" + \
          "      --description=TEXT             set description of the mirror to TEXT.\n" + \
          "      --title=TEXT                   set title for the mirror-log to TEXT.\n" + \
          "      --email=EMAIL                  set admin-email to EMAIL.\n" + \
          "      --log-template=FILE            template file for logfiles.\n" + \
          "      --index-template=FILE          template file for the index file.\n" + \
          "      --summary-template=FILE        template file for summary indizes.\n" + \
          "      --mail-user-cmd=COMMAND        command to send mail about changes.\n" + \
          "      --mail-users=FILE              mail info about changes to users in FILE.\n" + \
          "      --info-mailcmd=COMMAND         send mail to users when mirror changed.\n\n" + \
          "Download and FTP options:\n" + \
          "  -P, --parent=DIRECTORY         set parent directory for downloads to DIR.\n" + \
          "  -s, --parent-format=FORMAT     set parent directory format to FORMAT.\n" + \
          "  -D, --download-url=URL         set local download url to URL.\n" + \
          "  -t, --tries=NUMBER             set number of retries to NUMBER (0 unlimits).\n" + \
          "  -w, --wait=SECONDS             wait SECONDS between retries.\n" + \
          "  -r, --recursive                recursive download.\n" + \
          "  -F, --follow-symlinks          follow symbolic links.\n" + \
          "  -e, --exclude=REGEXP           set exclude pattern to REGEXP.\n" + \
          "  -i, --include=REGEXP           set include pattern to REGEXP.\n" + \
          "  -l, --only-latest              download only the latest file (matching -i).\n" + \
          "  -u, --username=TEXT            set username TEXT for downloads.\n" + \
          "  -p, --password=TEXT            set password TEXT for downloads.\n" + \
          "  -m, --max-parallel=NUMBER      set max. number of parallel downloads to NUM.\n" + \
          "  -R, --real-hostname=HOSTNAME   set real hostname to HOSTNAME (override URL).\n" + \
          "  -M, --max-days=NUMBER          set max. age of files to download to NUM. days.\n" + \
          "  -A, --max-age=NUMBER           set max. age of local files to NUMBER days.\n" + \
          "  -E, --max-delete=VALUE         delete max. VALUE files (KB, MB or percent).\n" + \
          "  -U, --umask=MASK               set the umask to MASK.\n" + \
          "      --ignore-mdtm              ignore mdtm when downloading files.\n" + \
          "      --owner=UID:GID            set the file owener to UID:GID.\n" + \
          "      --delete-local             always delete local files.\n" + \
          "      --passive                  set passive mode.\n" + \
          "      --continue                 restart getting an existing file.\n" + \
          "      --proxy=HOST:PORT          set proxy host and port.\n" + \
          "      --test                     set test mode (no real actions).\n\n" + \
          "Error handling options:\n" + \
          "      --error-mailcmd=COMMMAND   set error mailing command to COMMAND.\n\n" + \
          "Special debug options:\n" + \
          "      --debug                    debug mode (be very verbose).\n" + \
          "      --noexception              do not use exception handling.\n\n" + \
          "Advanced options:\n" + \
          "  -k, --no-kill             do not kill old processes.\n" + \
          "      --pid-file=FILENAME   set the pid-file to FILENAME.\n\n" + \
          "Please report any bugs/comments to Martin.Preishuber@eclipt.at\n"
PIRevLog = "Revision Log:\n\n" + \
         "    1.1b1 to b3      released as SimpleMirror in early 1998 \n" + \
         "    2.0b1 to b6      May - November 1998\n" + \
         "    2.0              February 9th, 1999\n" + \
         "    2.1pre2 to pre15 February 1999 - April 2000\n" + \
         "    2.1.16           August 21st, 2000\n" + \
         "    2.1.17           October 26th, 2000\n" + \
         "    2.1.18           April 18th, 2001\n" + \
         "    2.1.19           August 2nd, 2001\n" + \
         "    2.1.20           October 20th, 2001\n\n" + \
         "Have a look at the ChangeLog-file to get some more details.\n" + \
         "The most recent version can be found at:\n\n" + \
         "    http://www.eclipt.at   (Projects -> ECLiPt Mirroring Tool)\n" + \
         "    ftp://eclipt.uni-klu.ac.at/pub/projects/emirror\n\n" + \
         "Please report bugs, comments and feedback to Martin.Preishuber@eclipt.at\n"

# File Types (FT*)

FTDirectory = "Directory"
FTLink = "Link"
FTRegularFile = "Regular File"
FTLinkTarget = "LinkTarget"

# Log Messages (LM*)

LMError = "Error"
LMWarning = "Warning"
LMRetrieve = "Retrieve"
LMDelete = "Delete"
LMLink = "Link"

# Logfile Formats (LF*)

LFHTML = "html"
LFPlain = "plain"
LFPHP = "php"

# Logfile Extensions (LE*)

LEHTML = ".html"
LEPlain = ".txt"
LEPHP = ".php"

# Connection Types (CT*)

CTFTP = "ftp"
CTHTTP = "http"

# Connection Status

CSConnected = 1
CSDisconnected = 0

# Mirror status (MS*)

MSunchanged = "unchanged"
MSnotloggedin = "notloggedin"
MSunknown = "unknown"

# Default Values (DV*)

DVUsername = "ftp"
DVPassword = "emirror@"
DVRetries = 5
DVDelay = 30

# Parent Format (PF*)

PFFull = "full"
PFPath = "path"
PFNone = "none"

# Error Action (EA*)

EAContinue = "continue"
EAAbort = "abort"
EAIgnore = "ignore"
EARetry = "retry"

# Error Sources (ES*)

ESDownloadFile = "downloadfile"
ESGetRemoteFileListCwd = "getremotefilelist (cwd)"
ESGetRemoteFileListLS = "getremotefilelist (ls)"
ESDeleteFile = "DeleteFile"
ESGetLocalFileList = "GetLocalFileList"

# Installer constants (INS*)

INSBuildDir = "/var/tmp/"
INSRPMOpts = "-Uvh"
INSRPMBuildOpts = "--rebuild"
INSRPMDir = "/usr/src/redhat/RPMS"

# Miscellaneous constants

LastlogFilename = "lastlog.emirror"
EMirrorEnvironment = "EMIRRORCFGPATH"
EMirrorEnvOptions = "EMIRROROPTIONS"

MSDate = compile("\d\d-\d\d-\d\d")
MSMonthStrings = [ "jan", "feb", "mar", "apr", "may", "jun", "jul", "aug", "sep", "oct", "nov", "dec" ]



"""

The main program of EMirror

"""

# ----------------------------------------------------------------------
# Import functions I need 

from string import strip, find, split, lower, atoi, atof, atol, digits, rfind, capitalize, zfill, join
from sys import exit, argv, hexversion
from os import chdir, mkdir, utime, unlink, getcwd, chmod, symlink, lstat, readlink
from os import environ, rename, system, getuid, chown, remove, kill, getpid, umask
from time import time, mktime, sleep, ctime, localtime, asctime
if (hexversion != 0):  # FIXME: check this
    from ftplibbeta import FTP, all_errors, error_reply, error_temp, error_perm, error_proto
else:
    from ftplib import FTP, all_errors, error_reply, error_temp, error_perm, error_proto
from re import match, sub, compile
from types import TupleType, InstanceType, ListType
from urlparse import urlunparse
from tempfile import mktemp
from ConfigParser import ConfigParser
from error import errormsgs

from tools import *
from constants import *

import os
import getopt
import copy
import sys
import signal
import shutil
import socket
import popen2
import log4py
import process

# ----------------------------------------------------------------------

LOG_FMT = "%T %L %M"                              # LOG_FMT = log4py.FMT_DEBUG

# ----------------------------------------------------------------------
# Definition of the File-Class:

class File:
    
    def __init__(self):
        self.log4py = log4py.Logger().get_instance(self)

        self.name = ""                            # Name of the file
        self.filetype = None                      # FileType (File, Directory, Link)
        self.connection = None                    # Connection necessary for this file
        self.permissions = 0                      # File Permissions
        self.size = 0L                            # Size of the file

        self.linktarget = ""
        self.modtime = 0

    # split a file entry so that blanks are not lost, e.g. "a  b c.txt" -> ("a ", "b", "c.txt")
    def filesplit(self, string, character):
        list = []
        element = ""
        while (len(string) > 0):
            current = string[0]
            if (current == character):
                while (len(string) > 1) and (string[1] == character):
                    element = "%s%s" % (element, character)
                    string = string[1:]
                    current = string[0]
                list.append(element)
                element = ""
            else:
                element = "%s%s" % (element, current)
            string = string[1:]
        if (element != ""):
            list.append(element)
        return list

    def GetFilePerms(self, permstring):
        tempperms = 0

        # File permissions for owner
        if permstring[0] == "r":
            tempperms = tempperms | 256
        if permstring[1] == "w":
            tempperms = tempperms | 128
        if permstring[2] == "x":
            tempperms = tempperms | 64
        elif permstring[2] == "S":
            tempperms = tempperms | 2048
        elif permstring[2] == "s":
            tempperms = tempperms | 2112

        # File permissions for group
        if permstring[3] == "r":
            tempperms = tempperms | 32
        if permstring[4] == "w":
            tempperms = tempperms | 16
        if permstring[5] == "x":
            tempperms = tempperms | 8
        elif permstring[5] == "S":
            tempperms = tempperms | 1024
        elif permstring[5] == "s":
            tempperms = tempperms | 1032

        # File permissions for other
        if permstring[6] == "r":
            tempperms = tempperms | 4
        if permstring[7] == "w":
            tempperms = tempperms | 2
        if permstring[8] == "x":
            tempperms = tempperms | 1
        elif permstring[8] == "T":
            tempperms = tempperms | 512
        elif permstring[8] == "t":
            tempperms = tempperms | 513

        self.permissions = tempperms

    def SplitNonEPLF(self, filestruct):

        if (filestruct[0] == "-"):
            self.filetype = FTRegularFile
        elif (filestruct[0] == "d"):
            self.filetype = FTDirectory
        elif (filestruct[0] == "l"):
            self.filetype = FTLink
        permissionstring = filestruct[1:10]
        self.GetFilePerms(permissionstring)
        splitted = self.filesplit(filestruct, " ")

        self.name = ""
        if (lower(strip(splitted[5])) in MSMonthStrings):   # WU-FTPD < 2.6.*, ProFTPD and similar FTPD format
            self.size = atol(strip(splitted[4]))
            for i in range(8, len(splitted)):
                self.name = self.name + " " + splitted[i]
        else:                                               # WU-FTPD 2.6.* (some versions don't include the file owner)
            self.size = atol(strip(splitted[3]))
            for i in range(7, len(splitted)):
                self.name = self.name + " " + splitted[i]

        self.name = lstrip(self.name)

        if (self.filetype == FTLink):
            splitted = split(self.name, "->")
            self.name = strip(splitted[0])
            self.linktarget = strip(splitted[1])
        else:
            self.linktarget = ""

    def SplitEPLF(self, filestruct):
        splitted = split(filestruct[1:], ",")
        self.name = strip(splitted[len(splitted) - 1])
        self.linktarget = ""
        for i in range(len(splitted) - 1):
            parameter = strip(splitted[i])
            if (parameter == "/"):
                self.filetype = FTDirectory
                self.permissions = 493                               # rwxr-xr-x
            elif (parameter == "r"):
                self.filetype = FTRegularFile
                self.permissions = 420                               # rw-r--r--
            elif (parameter[0] == "s"):
                self.size = atol(parameter[1:])
            elif (parameter[0] == "m"):
                self.modtime = atoi(parameter[1:])

    def SplitMSFormat(self, filestruct):
        if (filestruct[24:29] == "<DIR>"):
            self.filetype = FTDirectory
            self.permissions = 493                                    # rwxr-xr-x
        else:
            self.filetype = FTRegularFile
            self.permissions = 420                                    # rw-r--r--
            self.size = atol(strip(filestruct[29:38]))
            
        self.name = filestruct[39:]

    def SplitFileInfo(self, filestruct, title):
        if (filestruct[0] == "+"):
            self.SplitEPLF(filestruct)
            self.log4py.debug("[ %s (debug) ] SplitFileInfo (%s): EPLF detected" % (str(title), self.name))
        elif (MSDate.match(filestruct)):
            self.SplitMSFormat(filestruct)
            self.log4py.debug("[ %s (debug) ] SplitFileInfo (%s): Microsoft (NT5 Beta ?) format detected" % (str(title), self.name))
        else:
            self.SplitNonEPLF(filestruct)
            self.log4py.debug("[ %s (debug) ] SplitFileInfo (%s): Standard unix format detected" % (str(title), self.name))

# ----------------------------------------------------------------------

class Installer:

    # Initialize the installer
    def __init__(self):
        self.index = 0                                      # The index of the source file (are ordered)
        self.title = None                                   # Mirror Title (= Installer Title)
        self.log4py = log4py.Logger().get_instance(self)
        self.sources = {}
        self.sourceregexps = {}
        self.srpms = []
        self.srpmregexp = None
        self.rpms = []
        self.rpmregexp = None
        self.rpmopts = INSRPMOpts
        self.rpmbuildopts = INSRPMBuildOpts
        self.builddir = INSBuildDir
        self.logfile = None                                 # Logfile of the build output

    # Rebuild a SRPM file
    def rebuildSRPM(self, filename):
        self.log4py.info("[ %s Installer ] Rebuilding SRPM %s" % (str(self.title), filename))
        rpms = ls(INSRPMDir, recursive = TRUE, includepath = TRUE)
        commandline = "rpm %s %s 2>&1" % (self.rpmbuildopts, filename)
        doit = popen2.popen2(commandline)
        output = doit[0].readlines()
        if (self.logfile != None):
            file = open(self.logfile, "a")
            file.write("\n[ %s Installer ] Rebuilding SRPM %s\n\n" % (str(self.title), filename))
            for i in range(len(output)):
                file.write(output[i])
            file.write("\n")
            file.close()
        newrpms = listdiff(rpms, ls(INSRPMDir, recursive = TRUE, includepath = TRUE))
        if (len(newrpms) > 0):
            self.rpms = self.rpms + newrpms
        else:
            self.log4py.error("[ %s Installer ] Rebuild of %s failed or RPM already exists. Check logfile for details." % (str(self.title), filename))

    # Install RPMs
    def installRPMs(self):
        filelist = ""
        for i in range(len(self.rpms)):
            filelist = "%s %s" % (filelist, self.rpms[i])
        self.log4py.info("[ %s Installer ] Installing RPMs: %s" % (str(self.title), strip(filelist)))

        commandline = "rpm %s %s 2>&1" % (self.rpmopts, strip(filelist))
        doit = popen2.popen2(commandline)
        output = doit[0].readlines()
        if (self.logfile != None):
            file = open(self.logfile, "a")
            file.write("\n[ %s Installer ] Installing RPMs: %s\n\n" % (str(self.title), strip(filelist)))
            for i in range(len(output)):
                file.write(output[i])
            file.write("\n")
            file.close()
        for i in range(len(output)):
            if find(output[i], "already installed") or find(output[i], "failed"):
                self.log4py.error("[ %s Installer ] Installation of RPMs failed. Check logfile for details." % str(self.title))

    # Install Source
    def installSource(self, file, commands):
        self.log4py.info("[ %s Installer ] Installing Source: %s [%s]" % (str(self.title), file, commands))

        if (find(commands, ";") != -1):
            splitted = split(commands, ";")
        else:
            splitted = [commands]

        currentdir = getcwd()
        chdir(self.builddir)

        dirlist = ls(self.builddir, includepath = TRUE, onlydirectories = TRUE)
        newdir = []
        filelog = open(self.logfile, "a")
        for i in range(len(splitted)):
            command = splitted[i]
            command = sub("%f", file, command)
            if (len(newdir) == 1):
                command = sub("%d", newdir[0], command)
            self.log4py.debug("[ %s Installer (debug) ] Executing: %s" % (str(self.title), command))
            filelog.write("\n[ %s Installer ] Source: Executing %s\n\n" % (str(self.title), command))
            if (command[:2] == "cd"):
                chdir(command[3:])
            else:
                doit = popen2.popen2("%s 2>&1" % command)
                output = doit[0].readlines()
                for i in range(len(output)):
                    filelog.write(output[i])
                filelog.write("\n")
            if (len(newdir) == 0):
                newdir = listdiff(dirlist, ls(self.builddir, includepath = TRUE, onlydirectories = TRUE))

        filelog.close()
        if (len(newdir) == 1):                              # Cleanup (remove build directory)
            rm(newdir[0])
        chdir(currentdir)

    # Do the real installation
    def finish(self):
        if (self.logfile != None) and (os.path.exists(self.logfile)): unlink(self.logfile)
        for i in range(len(self.srpms)):
            self.rebuildSRPM(self.srpms[i])
        if (len(self.rpms) > 0):
            self.installRPMs()
        keys = self.sources.keys()
        keys.sort()
        for i in range(len(keys)):
            key = keys[i]
            self.installSource(self.sources[key][0], self.sources[key][1])

    # Check wether a file has to be installed or not
    def checkfile(self, path, filename, localname):
        if (self.rpmregexp != None) and ((self.rpmregexp.match(filename)) or (self.rpmregexp.match(path + filename))):
            self.rpms.append(localname)
        if (self.srpmregexp != None) and ((self.srpmregexp.match(filename)) or (self.srpmregexp.match(path + filename))):
            self.srpms.append(localname)
        keys = self.sourceregexps.keys()
        keys.sort()
        for i in range(len(keys)):
            key = keys[i]
            if (self.sourceregexps[key][0].match(filename)) or (self.sourceregexps[key][0].match(path + filename)):
                self.sources[self.index] = [localname, self.sourceregexps[key][1]]
                self.index = self.index + 1

    # Add regular expressions to the installer (from the config file)
    def add(self, option, value):
        if (lower(option) == "rpmoptions"):
            self.rpmopts = value
        elif (lower(option) == "rpmbuildoptions"):
            self.rpmbuildopts = value
        elif (lower(option) == "logfile"):
            self.logfile = value
        elif (lower(option)[:4] == "rpms"):
            self.rpmregexp = compile(unfold(value))
        elif (lower(option)[:5] == "srpms"):
            self.srpmregexp = compile(unfold(value))
        elif (lower(option)[:5] == "other"):
            index = atoi(option[5:])
            filename = value[:find(value, "[") - 1]
            command = value[find(value, "[") + 1:-1]
            self.sourceregexps[index] = [compile(unfold(filename)), command]
        elif (lower(option) == "builddir"):
            self.builddir = value

# ----------------------------------------------------------------------

class Connection:
    def __init__(self):
        self.protocoll = None        # Protocoll to connect
        self.hostname = None         # Hostname of the remote host
        self.realhostname = ""       # real Hostname / IP (for hidden sites)
        self.port = None             # Port to connect to
        self.username = None         # Username
        self.password = None         # Password used ...
        self.status = None           # Connection status

# ----------------------------------------------------------------------
# Definition of the Mirror-Class:

class Mirror:

    # Initialisation of some variables
    def __init__(self):
        self.url = ""                             # Full URL of the mirror
        self.protocol = ""                        # Protocol to start with
        self.username = ""                        # Username
        self.password = ""                        # Password
        self.hostname = ""                        # Hostname / IP of the remote site
        self.realhostname = ""                    # real Hostname / IP (for hidden sites)
        self.path = ""                            # Remote directory to mirror
        self.directory = getcwd() + "/"           # Local directory
        self.noexception = FALSE                  # Do NOT use exceptions ?
        self.log4py = log4py.Logger().get_instance(self)
        self.test = FALSE                         # Test mode
        self.port = 0                             # Port Number
        self.delay = DVDelay                      # Delay time between retries

        self.followsymlinks = FALSE               # Do not follow symbolic links
        self.parentformat = PFFull                # Full parent directory format

        self.logdirectory = ""                    # Directory where the logfiles reside
        self.logfilename = None                   # Name of the logfile
        self.logfileformat = LFPlain              # Format of the logfiles (Plain, HTML)
        self.logdata = []                         # The Stuff which should appear in the log

        self.retries = DVRetries                  # Number of retries
        self.logfile = None                       # Filehandler of the logfile
        self.recursive = FALSE                    # Recursive download
        self.starttime = 0                        # Start time of the mirroring
        self.endtime = 0                          # End time of the mirroring
        self.category = "Assorted"                # Category of the mirror (for logfiles)
        self.description = None                   # Description of the mirror
        self.adminemail = None                    # EMail-adress of the ftp-maintainer
        self.errorcolor = "#CC0033"               # Default color for error messages
        self.warningcolor = "#191970"             # Default color for warning messages

        self.logtemplate = None                   # Template file for logs
        self.template = None                      # Template data
        self.templateloop = None                  # Template data for 

        self.maxdays = 0                          # maximum age of files to download (days)
        self.maxage = 0                           # maximum age of local files (days)
        self.maxdelete = None                     # maximum size or percantage of files to delete

        self.onlylatest = []                      # Only get the latest version
        self.title = None                         # Title of the Mirror
        self.localurl = None                      # Local download URL
        self.exclude = None                       # Exclude pattern
        self.include = None                       # Regular expression for file selection
        self.ignore = None                        # Regular expression to ignore local files
        self.deletelocal = FALSE                  # Always delete local files, which don't exist on the remote side
        self.errormailcmd = None                  # Command to execute on error
        self.infomailcmd = None                   # Command to execute when a mirror has changed
        self.ignoremdtm = FALSE                   # Do not ignore the MDTM when downloading files
        self.passivemode = FALSE                  # Use passive mode ?
        self.continueftp = FALSE                  # Continue FTP downloads ?
        self.umask = None

        self.notloggedin = FALSE                  # default: emirror could login

        self.filelist = []                        # List of files for iterative checking
        self.connection = None                    # Current connection
        self.ftp = FTP()                          # The FTP-Connection

        self.downloadfile = ""                    # Name of the file which is being downloaded
        self.currentfile = ""                     # the current file being processed

        self.remotefiles = []                     # Remote file listing for a new directory
        self.localfiles = []                      # Local file list for a new directory

        self.statfilelist = []                    # File-list for the MyStat-Function

        self.nrbytesretr = 0L                     # Number of bytes retrieved
        self.nrfilesretr = 0                      # Number of files retrieved
        self.nrdirscreated = 0                    # Number of directories created
        self.nrlinkscreated = 0                   # Number of links created
        self.nrbytesdel = 0L                      # Number of bytes deleted
        self.nrfilesdel = 0                       # Number of files deleted
        self.nrdirsdel = 0                        # Number of directories deleted
        self.nrlinksdel = 0                       # Number of links deleted

        self.owner = None                         # Set owner of downloaded files to this UID
        self.group = None                         # Set group of downloaded files to this GID

        self.filelistcache = {}                   # Simple File Listing cache
        self.installer = Installer()              # File Installer

        self.proxy = None        		  # Proxy host:port

    # Open the logfile ... that is notice the start time ;-)
    def StartLog(self):
        self.starttime = time()

    # Append a line to the logfile
    def AppendLog(self, line):
        if self.logfilename != None:

            if (line[0] == LMRetrieve) and (self.localurl != None):
                downloadurl = self.GetLCwd(parentdirectory = self.localurl, hostname = self.hostname) + self.currentfile.name
                downloadname = self.currentfile.path + self.currentfile.name
                if (self.logfileformat == LFPlain):
                    logrow = [ line[0], downloadname ]
                else:
                    logrow = [ line[0], "<A HREF=\"%s\">%s</A>" % (downloadurl, downloadname) ]
            else:
                logrow = [ line[0], line[1] ]

            self.logdata.append(logrow)

    # Read the logtemplate file
    def ReadLogTemplate(self):
        self.log4py.debug("[ %s (debug) ] Logfile: reading template file" % str(self.title))
        file = open(self.logtemplate, "r")
        line = "<empty>"
        self.template = []
        self.templateloop = []
        target = self.template

        if (self.logfileformat == LFHTML):
            dateiname = "index" + LEHTML
            fileupdatedname = "updated" + LEHTML
            fileerrorsname = "errors" + LEHTML
        elif (self.logfileformat == LFPHP):
            dateiname = "index" + LEPHP
            fileupdatedname = "updated" + LEPHP
            fileerrorsname = "errors" + LEPHP
        elif (self.logfileformat == LFPlain):
            dateiname = "index" + LEPlain
            fileupdatedname = "updated" + LEPlain
            fileerrorsname = "errors" + LEPlain

        while (line != ""):
            line = file.readline()
            line = sub("\$indexfilename\$", dateiname, line)
            line = sub("\$updatedfilename\$", fileupdatedname, line)
            line = sub("\$errorfilename\$", fileerrorsname, line)
            if (lower(strip(line)) == "$logstart$"):
                target.append("$logdata$")
                target = self.templateloop
            elif (lower(strip(line)) == "$logend$"):
                target = self.template
            elif (find(strip(lower(line)), "<!-- errorcolor:") != -1):
                color = strip(lower(line))
                position = find(color, "<!-- errorcolor:")
                color = color[position + len("<!-- errorcolor:"):]
                position = find(color, "-->")
                color = strip(color[:position])
                self.errorcolor = color
            elif (find(strip(lower(line)), "<!-- warningcolor:") != -1):
                color = strip(lower(line))
                position = find(color, "<!-- warningcolor:")
                color = color[position + len("<!-- warningcolor:"):]
                position = find(color, "-->")
                color = strip(color[:position])
                self.warningcolor = color
            else:
                target.append(line)
        file.close()

    # Check the logfilename and adopt its extension
    def CheckLogfilename(self, filename):
        if (rfind(filename, ".") != -1):
            extension = lower(filename[rfind(filename, "."):])
            basicname = filename[:rfind(filename, ".")]
        else:
            extension = ""
            basicname = filename
        if (extension == "") or (extension in [".html", ".asc", ".txt", ".php", ".php3", ".php4", ".htm"]):
            if (self.logfileformat == LFHTML):
                return basicname + LEHTML
            elif (self.logfileformat == LFPHP):
                return basicname + LEPHP
            elif (self.logfileformat == LFPlain):
                return basicname + LEPlain
        else:
            return filename
        
    # Write the whole log file at once
    def WriteLog(self):
        if self.logfilename != None:
            self.log4py.info("[ %s ] Logfile: writing" % str(self.title))

            if (not os.path.exists(self.logdirectory)):
                try:
                    mkdirtree(self.logdirectory)
                    self.log4py.info("[ %s ] Logfile: Directory \"%s\" created" % (str(self.title), self.logdirectory))
                except:
                    self.log4py.error("[ %s ] Logfile: Error: Couldn't create directory \"%s\" !" % (str(self.title), self.logdirectory))
                    return
            try:
                newFilename = self.CheckLogfilename(self.logfilename)
                self.logfile = open(self.logdirectory + newFilename, "w")
            except:
                self.log4py.error("[ %s ] Logfile: Error: Couldn't write logfile \"%s\" !" % (str(self.title), self.logfilename))
                return

            self.endtime = time()
            difference = long(self.endtime) - long(self.starttime)
            
            if (self.logtemplate == None):
                self.log4py.error("[ %s ] Logfile: Error: No template file specified !" % str(self.title))
                return

            # Read the template file
            self.ReadLogTemplate()

            # Unparse the URL for a nice outfit :)
            url = urlunparse((self.protocol, self.hostname, self.path, "", "", ""))

            # Check wether a local download-URL is set
            if (self.localurl != None):
                downloadurl = self.GetLCwd(hostname = self.hostname, path = self.path, parentdirectory = self.localurl)
            else:
                downloadurl = ""

            # Check the username and password
            if (lower(self.username) == "ftp") or (lower(self.username) == "anonymous"):
                username = self.username
                password = self.password
            else:
                username = "-- Hidden --"
                password = "-- Hidden --"

            # Check the administrator EMail address
            if self.adminemail != None:
                email = self.adminemail
            else:
                email = ""

            # Calculate the time
            totalseconds = difference
            minutes, seconds = divmod(totalseconds, 60)
            hours, minutes = divmod(minutes, 60)

            # Create Format specific text
            if (self.logfileformat == LFPlain):
                duration = "%2d Hour(s), %2d Minute(s) and %2d Second(s)" % (hours, minutes, seconds)
                downloaded = "%s Byte(s) in %.0f File(s), %.0f Link(s) and %.0f Directories ..." % (lng2str(self.nrbytesretr), self.nrfilesretr, self.nrlinkscreated, self.nrdirscreated)
                deleted = "%s Byte(s) in %.0f File(s), %.0f Link(s) and %.0f Directories ..." % (lng2str(self.nrbytesdel), self.nrfilesdel, self.nrlinksdel, self.nrdirsdel)
            else:
                duration = "<b>%2d</b> Hour(s), <b>%2d</b> Minute(s) and <b>%2d</b> Second(s)" % (hours, minutes, seconds)
                downloaded = "<b>%s</b> Byte(s) in <b>%.0f</b> File(s), <b>%.0f</b> Link(s) and <b>%.0f</b> Directories ..." % (lng2str(self.nrbytesretr), self.nrfilesretr, self.nrlinkscreated, self.nrdirscreated)
                deleted = "<b>%s</b> Byte(s) in <b>%.0f</b> File(s), <b>%.0f</b> Link(s) and <b>%.0f</b> Directories ..." % (lng2str(self.nrbytesdel), self.nrfilesdel, self.nrlinksdel, self.nrdirsdel)

            # Calculate download rate
            if (self.nrbytesretr > 0):
                if (totalseconds > 0):
                    rate = self.nrbytesretr / totalseconds
                else:
                    rate = 0
                kbytepersec = int(rate / 1024)
                if (kbytepersec > 0):
                    if (self.logfileformat == LFPlain):
                        downloadrate = "%.0f KByte/sec" % int(kbytepersec)
                    else:
                        downloadrate = "<b>%.0f</b> KByte/sec" % int(kbytepersec)
                else:
                    if (self.logfileformat == LFPlain):
                        downloadrate = "%.0f Byte/sec" % int(rate)
                    else:
                        downloadrate = "<b>%.0f</b> Byte/sec" % int(rate)
            else:
                downloadrate = "Nothing downloaded"

            # Calculate the entry for STATUS
            if (self.nrfilesretr == 0) and (self.nrfilesdel == 0) and (self.nrlinksdel == 0) and (self.nrdirscreated == 0) and (self.nrdirsdel == 0) and (self.nrlinkscreated == 0):
                if (self.notloggedin == FALSE):
                    status = MSunchanged
                    self.logdata.append([ "Information", "There's nothing new on the remote host" ])
                else:
                    status = MSnotloggedin
            else:
                status = str(self.endtime)

            statsstring = "%.0f %.0f %.0f %.0f %.0f %.0f %.0f %.0f %.0f %.0f" % (self.nrbytesretr, self.nrfilesretr, self.nrlinkscreated, self.nrdirscreated, self.nrbytesdel, self.nrfilesdel, self.nrlinksdel, self.nrdirsdel, self.starttime, self.endtime)

            # Process the template
            for i in range(len(self.template)):
                line = self.template[i]
                if (line == "$logdata$"):
                    for j in range(len(self.logdata)):
                        loginfo = self.logdata[j]
                        for k in range(len(self.templateloop)):
                            line = self.templateloop[k]
                            if (self.logfileformat != LFPlain) and (loginfo[0] == LMError):
                                line = sub("\$type\$", "<font color=\"%s\">%s</font>" % (self.errorcolor, loginfo[0]), line)
                                line = sub("\$message\$", "<font color=\"%s\">%s</font>" % (self.errorcolor, loginfo[1]), line)
                            elif (self.logfileformat != LFPlain) and (loginfo[0] == LMWarning):
                                line = sub("\$type\$", "<font color=\"%s\">%s</font>" % (self.warningcolor, loginfo[0]), line)
                                line = sub("\$message\$", "<font color=\"%s\">%s</font>" % (self.warningcolor, loginfo[1]), line)
                            else:
                                line = sub("\$type\$", loginfo[0], line)
                                line = sub("\$message\$", loginfo[1], line)
                            self.logfile.write(line)
                else:
                    line = sub("\$title\$", str(self.title), line)
                    line = sub("\$url\$", url, line)
                    line = sub("\$category\$", str(self.category), line)
                    line = sub("\$description\$", str(self.description), line)
                    line = sub("\$localurl\$", downloadurl, line)
                    line = sub("\$email\$", email, line)
                    line = sub("\$hostname\$", self.hostname, line)
                    line = sub("\$port\$", str(self.port), line)
                    line = sub("\$username\$", username, line)
                    line = sub("\$password\$", password, line)
                    line = sub("\$remotedirectory\$", self.path, line)
                    line = sub("\$localdirectory\$", self.directory, line)
                    line = sub("\$tries\$", str(self.retries), line)
                    line = sub("\$delay\$", str(self.delay), line)
                    line = sub("\$downloadrate\$", downloadrate, line)
                    line = sub("\$starttime\$", ctime(self.starttime), line)
                    line = sub("\$endtime\$", ctime(self.endtime), line)
                    line = sub("\$duration\$", duration, line)
                    line = sub("\$downloadstats\$", downloaded, line)
                    line = sub("\$deletestats\$", deleted, line)
                    line = sub("\$mdtm\$", status, line)
                    line = sub("\$statistics\$", statsstring, line)
                    line = sub("\$lastupdate\$", ctime(time()), line)
                    line = sub("\$timestamp\$", str(time()), line)
                    self.logfile.write(line)

            if (status != MSunchanged) and (status != MSnotloggedin) and (self.infomailcmd != None):
                self.log4py.info("[ %s ] Executing: %s" % (str(self.title), self.infomailcmd))
                if find(self.infomailcmd, "%s") != -1:
                    cmd = self.infomailcmd % str(self.title)
                else:
                    cmd = self.infomailcmd
                filename = mktemp()
                file = open(filename, "w")
                file.write("\nECLiPt-Mirror: Package \"%s\" updated\n\n" % str(self.title))
                file.write("[        Type ][ Message                                                      ]\n")
                for i in range(len(self.logdata)):
                    file.write("%s    %s\n" % (pfill(strip(self.logdata[i][0]), " ", 13), self.logdata[i][1]))
                file.write("\n")
                file.write("Generated at %s" % asctime(localtime(time())))
                file.close()
                system("cat %s | %s" % (filename, cmd))
                unlink(filename)
                
            self.logfile.close()
                
    # Check wether a local directory exists & create, if not
    def CheckDirectory(self, directory):

        if (directory[-1:] == "/"):
            directory = directory[:-1]
        if (os.path.exists(directory)) or (os.path.islink(directory)):
            if (os.path.islink(directory)) or (os.path.isfile(directory)):
                unlink(directory)
        if not os.path.exists(directory):
            pwd = getcwd()
            splitted = split(directory, "/")
            path = "/"
            for i in range(len(splitted)):
                if (os.path.exists(path + splitted[i])) and (os.path.isfile(path + splitted[i])):
                    unlink(path + splitted[i])
                if (not os.path.exists(path + splitted[i])) and (not os.path.islink(path + splitted[i])):
                    try:
                        chdir(path)
                    except:
                        self.log4py.error("[ %s ] Error: couldn't change to directory %s" % (str(self.title), directory))
                        exit(1)
                    try:
                        mkdir(splitted[i])
                        if (self.owner != None):
                            chown(splitted[i], self.owner, self.group)
                    except:
                        self.log4py.error("[ %s ] Error: Couldn't create directory %s" % (str(self.title), directory))
                        exit(1)
                    self.nrdirscreated = self.nrdirscreated + 1
                path = path + splitted[i] + "/"
            chdir(pwd)

    # Create the correct linktargetname and path from (name, path, target)
    def AdoptLinkTarget(self, name, path, target):

        if (target[0] != "/"):                              # it's a relative link -> adopt the linktarget
            newpath = path
            newname = target
            if newpath[-1:] != "/":
                newpath = newpath + "/"
            if (newname[0:3] == "../"):
                newpath = newpath[:-1]
                while (newname[0:3] == "../"):
                    newname = newname[3:]
                    newpath = newpath[0:rfind(newpath, "/")]
                newpath = newpath + "/"
                if (rfind(newname, "/") != -1):
                    newpath = newpath + newname[:rfind(newname, "/") + 1]
                    newname = newname[rfind(newname, "/") + 1:]
            while (find(newname, "/") != -1):
                newpath = newpath + newname[:find(newname, "/") + 1]
                newname = newname[find(newname, "/") + 1:]
        else:                                               # the linktarget is a absolut link
            newpath = target
            newname = ""

        return (newname, newpath)
    
    # Create a symbolic link
    def CreateLink(self):

        self.currentfile.linktarget = sub("//", "/", self.currentfile.linktarget)

        createlink = TRUE                                        # Default: create link
        append = TRUE                                            # Append file to the list (only if followsymlink is true)

        newFile = File()
        newFile.log4py.set_loglevel(self.log4py.get_loglevel())
        newFile.filetype = FTLinkTarget

        # Check the Linktarget and modify name & path if necessary

        newFile.name, newFile.path = self.AdoptLinkTarget(self.currentfile.name, self.currentfile.path, self.currentfile.linktarget)
        
        newFile.connection = self.currentfile.connection

        # Check Include / Exclude rules
        
        if (self.parentformat == PFNone) and ((len(newFile.path) < len(self.path)) or (newFile.path[:len(self.path)] != self.path)):
            self.log4py.warn("[ %s ] Warning: Link %s%s is outside the parent directory !" % (str(self.title), newFile.path, newFile.name))
            createlink = FALSE
        else:

            # Include rules
            if (self.include != None) and ((self.include.match(newFile.name) == None) and (self.include.match(newFile.path + newFile.name) == None)):
                createlink = FALSE

            # Exclude rules
            if ((self.exclude != None) and ((self.exclude.match(newFile.name) != None) or self.exclude.match(newFile.path + newFile.name) != None)):
                createlink = FALSE

            # Ignore rules
            if ((self.ignore != None) and ((self.ignore.match(newFile.name) != None) or self.ignore.match(newFile.path + newFile.name) != None)):
                createlink = FALSE

            if ((self.currentfile.path + self.currentfile.name) == (newFile.path + newFile.name)): # Check wether the link points to itself (endless loop)
                self.log4py.warn("[ %s ] Warning: Link %s%s points to itself !" % (str(self.title), self.currentfile.path, self.currentfile.name))
                append = FALSE

            if (newFile.name == "."):                                 # Check wether the link points to the current directory (endless loop)
                self.log4py.warn("[ %s ] Warning: Link %s%s points to . (current directory) !" % (str(self.title), self.currentfile.path, self.currentfile.name))
                createlink = FALSE

        if (self.followsymlinks == TRUE) and (createlink == TRUE) and (append == TRUE): # Don't append the file if you don't want to follow links
            self.filelist.append(newFile)

        pwd = getcwd()
        directory = self.GetLCwd()
        chdir(directory)

        deletelink = FALSE
            
        if os.path.islink(self.currentfile.name):                     # check wether the link already exists
            linktarget = readlink(self.currentfile.name)
            if (linktarget != self.currentfile.linktarget):           # if the linktarget has changed -> delete the link
                deletelink = TRUE
            if (createlink == FALSE):                                 # if the link exists, but shouldn't -> delete it
                deletelink = TRUE                                     # This makes the mirror clean (hopefully without dead links)

        elif os.path.exists(self.currentfile.name):       	      # this is, if a directory or file has become a link ... delete the old stuff then
            deletelink = TRUE

        if (os.path.islink(self.currentfile.name) and (createlink == FALSE)) or (deletelink == TRUE): # The link exists, but it shouldn't
            self.log4py.info("[ %s ] Deleting Link / File / Directory (FTP): %s%s " % (str(self.title), self.currentfile.path, self.currentfile.name))
            result = rm(self.currentfile.name)
            self.AppendLog([LMDelete, self.currentfile.path + self.currentfile.name])
            self.nrdirsdel = self.nrdirsdel + result[0]
            self.nrfilesdel = self.nrfilesdel + result[1]
            self.nrbytesdel = self.nrbytesdel + result[2]
            self.nrlinksdel = self.nrlinksdel + result[3]
            
        if (createlink == TRUE) and (not os.path.islink(self.currentfile.name)):
            self.log4py.info("[ %s ] Creating Link (FTP): %s%s -> %s" % (str(self.title), self.currentfile.path, self.currentfile.name, self.currentfile.linktarget))
            
            try:
                symlink(self.currentfile.linktarget, self.currentfile.name)
            except os.error, detail:
                self.log4py.warn("[ %s ] Warning: couldn't create link %s%s -> %s: %s" % (
                    str(self.title), self.currentfile.path, self.currentfile.name, self.currentfile.linktarget, detail))
            self.AppendLog([LMLink, self.currentfile.path + self.currentfile.name + " -> " + self.currentfile.linktarget])
            self.nrlinkscreated = self.nrlinkscreated + 1

        chdir(pwd)

    # Get Remote Files
    def GetRemoteFiles(self):
        self.log4py.debug("[ %s (debug) ] Getting list of remote files" % str(self.title))
        files2download = self.remotefiles
        for i in range(len(files2download)):
            remotefile = files2download[i]
            download = FALSE
            if (remotefile.filetype == FTDirectory):
                if (self.recursive == TRUE):
                    download = TRUE
            elif (remotefile.filetype == FTRegularFile) or (remotefile.filetype == FTLink):
                download = TRUE

            if (download == TRUE):
                # Insert file at the beginning to reduze memory usage
                if (remotefile.filetype == FTLink) or (remotefile.filetype == FTRegularFile):
                    self.filelist.insert(0, remotefile)
                else:
                    self.filelist.append(remotefile)

    # Delete a file or subtree
    def DeleteFile(self, path, directory, filename, reason = ""):

        # Do not delete which are being ignored !
        if (self.ignore != None) and ((self.ignore.match(filename) != None) or (self.ignore.match(directory + filename) != None)):
            return
        else:

            if (self.maxdelete > -1):
                if (self.nrbytesdel < self.maxdelete):
                    if (self.test == FALSE):
                        self.log4py.info("[ %s ] Deleting: %s%s (max %s Byte) %s" % (str(self.title), path, filename, lng2str(self.maxdelete), reason))
                    else:
                        self.log4py.info("[ %s (Test) ] Deleting: %s%s (max %s Byte) %s" % (str(self.title), path, filename, lng2str(self.maxdelete), reason))                            
            else:
                if (self.test == FALSE):
                    self.log4py.info("[ %s ] Deleting: %s%s %s" % (str(self.title), path, filename, reason))
                else:
                    self.log4py.info("[ %s (Test) ] Deleting: %s%s %s" % (str(self.title), path, filename, reason))
            try:
                if (self.nrbytesdel < self.maxdelete) or (self.maxdelete <= -1):
                    if (self.test == FALSE):
                        result = rm(directory + filename, self.maxdelete, self.nrbytesdel)
                    else:
                        result = [0, 0, 0, 0]
                else:
                    result = [0, 0, 0, 0]
            except OSError, detail:
                self.HandleError(detail, ESDeleteFile, extra = "filename / directory: \"%s/%s\")" % (directory, filename))
                return FALSE
            
            self.AppendLog([LMDelete + " " + reason,  path + filename])
            self.nrdirsdel = self.nrdirsdel + result[0]
            self.nrfilesdel = self.nrfilesdel + result[1]
            self.nrbytesdel = self.nrbytesdel + result[2]
            self.nrlinksdel = self.nrlinksdel + result[3]

    # Delete local files, which don't exist on the remote site any more 
    def DelOldLocalFiles(self):
        self.log4py.debug("[ %s (debug) ] Deleting old local files" % str(self.title))
        directory = self.GetLCwd()
        if (self.maxage != 0):
            maxmdtm = long(time() - (self.maxage * 60 * 60 * 24))

        localfilenames = []                                          # Remove files which are older than maxage days
        for i in range(len(self.localfiles)):
            if (self.maxage != 0):
                filename = directory + "/" + self.localfiles[i]
                if (os.path.isfile(filename) and (not os.path.islink(filename))):
                    if (lstat(filename)[8] < maxmdtm):
                        self.DeleteFile(self.currentfile.path, directory, self.localfiles[i], "(file older than " + str(self.maxage) + " day(s))")
                    else:
                        localfilenames.append(self.localfiles[i])
                else:
                    localfilenames.append(self.localfiles[i])
            else:
                localfilenames.append(self.localfiles[i])

        remotefilenames = []
        for i in range(len(self.remotefiles)):
            if (self.remotefiles[i].filetype == FTDirectory):
                remotefilenames.append(self.remotefiles[i].path[rfind(self.remotefiles[i].path, "/", 0, -2) + 1:][:-1])
            else:
                remotefilenames.append(self.remotefiles[i].name)
        remotefilenames.sort()

        for i in range(len(localfilenames)):
            filename = localfilenames[i]
            
            if (not filename in remotefilenames):
                if (not (self.IsTempFilename(filename) and (self.RealFilename(filename) in remotefilenames))) or (self.continueftp == FALSE):
                    if (self.include == None) or (self.deletelocal == TRUE) or ((self.include != None) and ((self.include.match(filename) != None) or (self.include.match(self.currentfile.path + filename) != None))):
                        self.DeleteFile(self.currentfile.path, directory, filename)

    # Get the local filelist
    def GetLocalFileList(self):
        self.log4py.debug("[ %s (debug) ] Getting list of local files" % (str(self.title)))
        directory = self.GetLCwd()
        try:
            self.localfiles = listdir(directory)
        except OSError, detail:
            self.localfiles = []
            self.HandleError(detail, ESGetLocalFileList, extra = "listdir(\"%s\")" % directory)

    # Callback for listing files
    def LsCallBack(self, line):
        if (line != "") and (lower(line) != "ls: .: permission denied") and (line[0] in [ "-", "d", "l", "+" ]) or (MSDate.match(line)):
            newfile = File()
            newfile.log4py.set_loglevel(self.log4py.get_loglevel())
            newfile.SplitFileInfo(line, self.title)
            if (newfile.name != ".") and (newfile.name != ".."):
                self.remotefiles.append(newfile)

    # Get the remote filelist
    def GetRemoteFileList(self, directory = "", cache = TRUE, exclude = TRUE):
        if (directory == ""):
            directory = self.currentfile.path
        if self.filelistcache.has_key(directory) and (cache == TRUE):
            self.log4py.info("[ %s ] Get Filelist: %s (cached)" %(str(self.title), directory))
            self.remotefiles = copy.copy(self.filelistcache[directory])
        else:
            self.log4py.info("[ %s ] Get Filelist: %s" % (str(self.title), directory))

            if (exclude == FALSE):
                tempinclude = self.include
                tempexclude = self.exclude
                self.include = self.exclude = None

            self.remotefiles = []

            # Change to the directory
            if (self.noexception == TRUE):
                self.ftp.cwd(directory)
            else:
                try:
                    self.ftp.cwd(directory)
                except all_errors, detail:
                    self.HandleError(detail, ESGetRemoteFileListCwd, extra = "ftp.cwd(\"%s\")" % directory)
                    return FALSE

            # Get the filelist
            if (self.noexception == TRUE):
                response = self.ftp.retrlines("LIST -a", self.LsCallBack)
            else:
                try:
                    # self.ftp.dir(self.LsCallBack)                  # This doesn't receive hidden directories
                    response = self.ftp.retrlines("LIST -a", self.LsCallBack)
                except (error_reply, error_temp, error_perm, error_proto, socket.error, IOError), detail:
                    self.HandleError(detail, ESGetRemoteFileListLS, extra = "ftp.dir(\"%s\")" % directory)
                    return FALSE
                except EOFError, detail:
                    self.HandleError("901 ftplib internal: connection lost", ESGetRemoteFileListLS, extra = "ftp.dir(\"%s\")" % directory)
                    return FALSE

            for i in range(len(self.remotefiles)):                   # Fixup name and path of all files
                if (self.remotefiles[i].filetype == FTDirectory):
                    self.remotefiles[i].path = directory + self.remotefiles[i].name + "/"
                    self.remotefiles[i].name = ""
                else:
                    self.remotefiles[i].path = directory
                self.remotefiles[i].connection = self.currentfile.connection

            newremotefiles = []                                      # Remove files, which don't match include or exclude
            for i in range(len(self.remotefiles)):

                remove = FALSE

                if (self.remotefiles[i].filetype == FTDirectory):
                    fullname = self.remotefiles[i].path[:-1]
                    filename = fullname[rfind(fullname, "/") + 1:]
                else:
                    filename = self.remotefiles[i].name
                    fullname = directory + filename

                # Rules for including files
                if (self.include != None) and ((self.include.match(filename) == None) and (self.include.match(fullname) == None)):
                    remove = TRUE

                # Rules for excluding files
                if (self.exclude != None) and ((self.exclude.match(filename) != None) or (self.exclude.match(fullname) != None)):
                    remove = TRUE

                # Rules for ignoreing files
                if (self.ignore != None) and ((self.ignore.match(filename) != None) or (self.ignore.match(fullname) != None)):
                    remove = TRUE
                
                if (remove == FALSE):
                    newremotefiles.append(self.remotefiles[i])

            self.remotefiles = newremotefiles
            
            if (exclude == FALSE):
                self.include = tempinclude
                self.exclude = tempexclude

            if (cache == TRUE):
                self.filelistcache[directory] = copy.copy(self.remotefiles)

    # Do some error handling
    def HandleError(self, error, source, extra = None):
        if (str(error) != ""):
            if (type(error) == TupleType) or (type(error) == InstanceType):
                errornumber = atoi(str(error[0]))
                errormessage = str(error[1])
            else:
                errornumber = atoi(str(error)[0:3])
                errormessage = strip(str(error)[4:])
        else:
            errornumber = 999
            errormessage = "emirror internal: unknown error"
        self.log4py.error("[ %s ] Error: #%d - %s" % (str(self.title), errornumber, errormessage)) 
        found = FALSE
        for i in range(len(errormsgs)):
            oneerror = errormsgs[i]
            if (errornumber == oneerror[0]) and (find(lower(errormessage), lower(oneerror[1])) != -1):
                found = TRUE
                if (oneerror[2] == EAContinue):
                    self.AppendLog([LMError, self.currentfile.path + self.currentfile.name + ": " + str(errornumber) + "-" + errormessage])
                elif (oneerror[2] == EAAbort):
                    self.End(1)
                elif (oneerror[2] == EARetry):
                    self.ReconnectFTP()

        if (found == FALSE):
            self.log4py.error("%s" % PIHeader)
            self.log4py.error("          Path: %s" % self.currentfile.path)
            self.log4py.error("          Name: %s" % self.currentfile.name)
            self.log4py.error("        Source: %s" % str(source))
            self.log4py.error("         Error: %s" % str(error))
            self.log4py.error("   Type(Error): %s" % str(type(error)))
            if (type(error) == InstanceType):
                self.log4py.error("    Dictionary: %s" % str(error.__dict__))
            self.log4py.error("   Errornumber: %d" % errornumber)
            self.log4py.error("  Errormessage: %s" % errormessage)
            self.log4py.error("         Extra: %s\n" % str(extra))

            if (self.errormailcmd != None):
                self.log4py.info("Executing: %s" % self.errormailcmd)
                filename = mktemp()
                file = open(filename, "w")
                file.write("%s" % PIHeader)
                file.write("\nError Report (%s):\n\n" % str(self.title))
                file.write("   Admin EMail: %s\n" % str(self.adminemail))
                file.write("    Mirror URL: %s\n" % str(self.url))
                file.write("          Path: %s\n" % self.currentfile.path)
                file.write("          Name: %s\n" % self.currentfile.name)
                file.write("        Source: %s\n" % str(source))
                file.write("         Error: %s\n" % str(error))
                file.write("   Type(Error): %s\n" % str(type(error)))
                if (type(error) == InstanceType):
                    file.write("    Dictionary: %s\n" % str(error.__dict__))
                file.write("   Errornumber: %d\n" % errornumber)
                file.write("  Errormessage: %s\n" % errormessage)
                file.write("         Extra: %s\n\n" % str(extra))
                file.write("Please try and correct this error as soon as possible.\n\n")
                file.close()
                system("cat %s | %s" % (filename, self.errormailcmd))
                unlink(filename)
            else:
                self.log4py.error("Please report this error to Martin.Preishuber@eclipt.at")
                self.log4py.error("or check error.py and adapt it (it's not really difficult ;-))")

            self.AppendLog([LMError, str(errornumber) + "-" + errormessage])
            self.WriteLog()
            exit(1)

    # Connect to a Website
    def ConnectHTTP(self):
        self.log4py.warn("[ %s ] Connecting (HTTP): not supported yet (use wget meanwhile)" % str(self.title))

    # Disconnect from a Website
    def DisconnectHTTP(self):
        self.log4py.warn("[ %s ] Disconnecting (HTTP): not supported yet (use wget meanwhile)" % str(self.title))

    # Connect to a FTP-site
    def ConnectFTP(self):
        if self.connection.realhostname != "":
            hostname = self.connection.realhostname
        else:
            hostname = self.connection.hostname
        self.log4py.info("[ %s ] Connecting (FTP): %s port %d" % (str(self.title), hostname, self.connection.port))
        nrretries = 1
        errormessage = None
        successful = FALSE
        if (self.retries == 0):
            unlimited = TRUE
        else:
            unlimited = FALSE
        while (((nrretries <= self.retries) or (unlimited == TRUE)) and (successful == FALSE)):
            if (self.proxy == None):
                if (nrretries == 1):
                    self.log4py.info("[ %s ] FTP login: first try (out of %d)" % (str(self.title), self.retries))
                else:
                    self.log4py.info("[ %s ] FTP login: retry #%d (out of %d) [%s]" % (str(self.title), nrretries - 1, self.retries, str(errormessage)))
            elif (self.proxy != None):
                if (nrretries == 1):
                    self.log4py.info("[ %s ] FTP login (proxy: %s): first try (out of %d)" % (str(self.title), self.proxy, self.retries))
                else:
                    self.log4py.info("[ %s ] FTP login (proxy: %s): retry #%d (out of %d) [%s]" % (str(self.title), self.proxy, nrretries - 1, self.retries, str(errormessage)))
 
            try:

                # Connect via a Proxy host
                if (self.proxy != None):
                    proxyhost, proxyport = split(self.proxy, ":")
                    self.ftp.connect(proxyhost, atoi(proxyport))
                    self.ftp.login("%s@%s:%d" % (self.connection.username, hostname, self.connection.port), self.connection.password)
                    successful = TRUE

                # Connect directly
                else:
                    self.ftp.connect(hostname, self.connection.port)
                    self.ftp.login(self.connection.username, self.connection.password)
                    successful = TRUE
            except all_errors, detail:
                if (type(detail) == ListType):
                    self.log4py.debug("[ %s (Debug) ] FTP login failed: %s - %s" % (str(self.title), detail[0], detail[1]))
                else:
                    self.log4py.debug("[ %s (Debug) ] FTP login failed: %s" % (str(self.title), detail))
                nrretries = nrretries + 1
                sleep(self.delay)
                errormessage = detail

        if (successful != TRUE):
            self.log4py.warn("[ %s ] FTP login failed after %d tries (%s)" % (str(self.title), self.retries, str(errormessage)))
            self.AppendLog([LMError, "Couldn't login to \"%s\" after %d tries (%s)" % (self.connection.hostname, self.retries, str(errormessage))])
            self.notloggedin = TRUE
        else:
            if (self.passivemode == TRUE):
                self.log4py.info("[ %s ] FTP: Setting passive mode" % str(self.title))
                self.ftp.set_pasv(1)
            else:
                self.ftp.set_pasv(0)
            errormessage = None
            self.log4py.info("[ %s ] FTP login succeeded" % str(self.title))
        return errormessage

    # Disconnect from a FTP-site
    def DisconnectFTP(self):
        try:
            self.ftp.quit()
        except:
            pass                                # no need to worry when disconnection failed

    # Reconnect from/to a FTP-site
    def ReconnectFTP(self):
        try:
            self.ftp.quit()
            self.log4py.info("[ %s ] FTP Reconnect: disconnect succeeded" % str(self.title))
        except:
            self.log4py.info("[ %s ] FTP Reconnect: disconnect failed" % str(self.title))
        result = self.ConnectFTP()
        if (result != None):
            self.log4py.error("[ %s ] FTP Reconnect: failed" % str(self.title))
            exit(1)

        self.filelist.insert(0, self.currentfile)

    # Close a connection (ftp or http)
    def CloseConnection(self):
        if (self.connection != None) and (self.connection.status == CSConnected):
            if (self.connection.protocol == CTFTP):
                self.DisconnectFTP()
            elif (self.connection.protocol == CTHTTP):
                self.DisconnectHTTP()
            self.connection.status = CSDisconnected
                
    # Open a connection (ftp or http)
    def OpenConnection(self):
        if (self.connection.protocol == CTFTP):
            result = self.ConnectFTP()
        elif (self.connection.protocol == CTHTTP):
            result = self.ConnectHTTP()
        if (result == None):
            self.connection.status = CSConnected
        return result

    # Callback for downloading a file
    def RetrCallBack(self, line):
        self.downloadfile.write(line)

    # Get the modification time of a remote file
    def GetRemoteMDTM(self, file):
        if (file.modtime == 0):
            try:
                modtime = split(self.ftp.voidcmd("MDTM " + file.path + file.name), " ")[1]
                if (modtime[0:3] == "191"):                 # Fix for old WU-FTP Daemons (y2k bug)
                    modtime = "2000%s" % (modtime[5:])
                year = atoi(modtime[0:4])
                month = atoi(modtime[4:6])
                day = atoi(modtime[6:8])
                hours = atoi(modtime[8:10])
                minutes = atoi(modtime[10:12])
                seconds = atoi(modtime[12:14])
                file.modtime = int(mktime((year, month, day, hours, minutes, seconds, 0, 0, -1)))
            except:
                file.modtime = -1

    # Returns the download filename for a given filename
    def TempFilename(self, filename):
        return "%s/.%s.tmp" % (os.path.dirname(filename), os.path.basename(filename))

    # Checks wether a filename is a tempfile or not
    def IsTempFilename(self, filename):
        tmpfilename = os.path.basename(filename)
        if len(tmpfilename) >= 5:
            if (tmpfilename[0] == ".") and (tmpfilename[-4:] == ".tmp"):
                return TRUE
            else:
                return FALSE
        else:
            return FALSE

    # Returns the real filename of a tempfile
    def RealFilename(self, filename):
        if len(filename) >= 5:
            return filename[1:-4]
        else:
            return filename
        
    # The real file-downloading function (FTP)
    def DownloadFileFTP(self, localfilepath, filename):
        rest = None
        tmpfilename = self.TempFilename(filename)
        if os.path.exists(tmpfilename):
            if (self.continueftp == TRUE):
                size = lstat(tmpfilename)[6]                          # determine file size
                if (self.test == FALSE):
                    self.log4py.info("[ %s ] Restarting at position %.0f (FTP): %s%s" % (str(self.title), size, self.currentfile.path, self.currentfile.name))
                    self.downloadfile = open(tmpfilename, "a")
                    rest = size
                else:
                    self.log4py.info("[ %s (Test) ] Restarting at position %.0f (FTP): %s%s" % (str(self.title), size, self.currentfile.path, self.currentfile.name))
            else:
                if (self.test == FALSE):
                    unlink(tmpfilename)
                    self.downloadfile = open(tmpfilename, "w")
        else:
            if (self.test == FALSE):
                self.downloadfile = open(tmpfilename, "w")

        if (self.test == FALSE):
            self.ftp.retrbinary("RETR %s%s" % (self.currentfile.path, self.currentfile.name), self.RetrCallBack, 1024, rest)
            self.downloadfile.close()
            rename(tmpfilename, filename)
            if (self.owner != None):
                chown(filename, self.owner, self.group)
            if (self.currentfile.modtime != -1):
                utime(localfilepath + self.currentfile.name, (self.currentfile.modtime, self.currentfile.modtime))
            if (self.umask == None):
                chmod(localfilepath + self.currentfile.name, self.currentfile.permissions)

        self.AppendLog([LMRetrieve, "%s%s" % (self.currentfile.path, self.currentfile.name)])

        self.nrbytesretr = self.nrbytesretr + self.currentfile.size
        self.nrfilesretr = self.nrfilesretr + 1
        self.installer.checkfile(self.currentfile.path, self.currentfile.name, filename)
    
    # Download a file
    def DownloadFile(self):
        localfilepath = self.GetLCwd()
        downloadfile = FALSE
        setmdtm = FALSE

        if (self.currentfile.modtime == 0):
            self.GetRemoteMDTM(self.currentfile)
        if (not (os.path.exists(localfilepath + self.currentfile.name))):
            downloadfile = TRUE
        else:
            localinfo = lstat(localfilepath + self.currentfile.name)
            if ((self.currentfile.modtime != int(localinfo[8]) and (self.currentfile.modtime != -1) and (self.ignoremdtm == FALSE))) or (self.currentfile.size != localinfo[6]):
                rm(localfilepath + self.currentfile.name)
                downloadfile = TRUE
            elif ((self.ignoremdtm == TRUE) and (self.currentfile.modtime != -1) and (self.currentfile.modtime != int(localinfo[8]))):
                setmdtm = TRUE

        # If local files older than maxage are been deleted then there's no need to download those
        if (downloadfile == TRUE) and (self.currentfile.modtime != -1) and (self.maxage != 0):
            maxmdtm = long(time() - (self.maxage * 60 * 60 * 24))
            if (self.currentfile.modtime < maxmdtm):
                downloadfile = FALSE

        # Don't download files newer than maxdays
        if (downloadfile == TRUE) and (self.currentfile.modtime != -1) and (self.maxdays != 0):
            newmdtm = long(time() - (self.maxdays * 60 * 60 * 24))
            if (self.currentfile.modtime < newmdtm):
                downloadfile = FALSE

        if (downloadfile == TRUE):
            if (self.test == FALSE):
                self.log4py.info("[ %s ] Downloading (FTP): %s%s" % (str(self.title), self.currentfile.path, self.currentfile.name))
            else:
                self.log4py.info("[ %s (Test) ] Downloading (FTP): %s%s" % (str(self.title), self.currentfile.path, self.currentfile.name))                   
            filename = "%s%s" % (self.GetLCwd(), self.currentfile.name)
            if (self.noexception == TRUE):
                self.DownloadFileFTP(localfilepath, filename)
            else:
                try:
                    self.DownloadFileFTP(localfilepath, filename)
                except all_errors, detail:
                    self.HandleError(detail, ESDownloadFile)
        elif ((setmdtm == TRUE) and (self.currentfile.modtime != -1)):
            if (self.test == FALSE):
                self.log4py.info("[ %s ] Setting MDTM (FTP): %s%s" % (str(self.title), self.currentfile.path, self.currentfile.name))
            else:
                self.log4py.info("[ %s (Test)] Setting MDTM (FTP): %s%s" % (str(self.title), self.currentfile.path, self.currentfile.name))
            try:
                if (self.test == FALSE):
                    utime(localfilepath + self.currentfile.name, (self.currentfile.modtime, self.currentfile.modtime))
            except all_errors, detail:
                self.HandleError(detail, ESDownloadFile)
            
    # Optimizes the remote filelist, so that only the latest version will be retrieved
    def ReduceRemoteFileList(self):
        allremotefiles = []

        # Check the remote filelist for files matching each only-latest regexp
        for i in range(len(self.onlylatest)):
            regexpremotefiles = []
            for j in range(len(self.remotefiles)):
                if ((self.onlylatest[i].match(self.remotefiles[j].name) != None) or (self.onlylatest[i].match(self.remotefiles[j].path + self.remotefiles[j].name))) and (self.remotefiles[j].filetype != FTDirectory):
                    regexpremotefiles.append(self.remotefiles[j])
            if (len(regexpremotefiles) > 1):
                mostrecentfile = File()
                for j in range(len(regexpremotefiles)):
                    self.GetRemoteMDTM(regexpremotefiles[j])
                    if (regexpremotefiles[j].modtime > mostrecentfile.modtime):
                        mostrecentfile = regexpremotefiles[j]
                allremotefiles.append(mostrecentfile)
            elif (len(regexpremotefiles) == 1):
                allremotefiles.append(regexpremotefiles[0])

        # Append all directories and files which do not match _any_ only-latest regexp
        for i in range(len(self.remotefiles)):
            if (self.remotefiles[i].filetype == FTDirectory):
                allremotefiles.append(self.remotefiles[i])
            else:
                append = TRUE
                for j in range(len(self.onlylatest)):
                    if ((self.onlylatest[j].match(self.remotefiles[i].name)) != None) or (self.onlylatest[j].match(self.remotefiles[i].path + self.remotefiles[i].name) != None):
                        append = FALSE
                if (append == TRUE):
                    allremotefiles.append(self.remotefiles[i])

        self.remotefiles = allremotefiles
        
    # Get the filetype of the first file (the one specified on the command line)
    def GetFileType(self, cache = TRUE, exclude = TRUE, firstURL = FALSE):
        self.currentfile.filetype = None
        if (self.currentfile.name == "") and (self.currentfile.path == "/"):
            self.currentfile.filetype = FTDirectory
        else:
            if (self.currentfile.name == ""):
                if (self.currentfile.path[-1] == "/") and (firstURL == TRUE):
                    try:
                        self.ftp.cwd(self.currentfile.path)
                        self.currentfile.filetype = FTDirectory
                    except:
                        directory = self.currentfile.path[:rfind(self.currentfile.path, "/", 0, -2)] + "/"
                else:
                        directory = self.currentfile.path[:rfind(self.currentfile.path, "/", 0, -2)] + "/"
            else:
                directory = self.currentfile.path
            if (self.currentfile.filetype == None):
                result = self.GetRemoteFileList(directory, cache, exclude)
                if (result == None):
                    for i in range(len(self.remotefiles)):
                        if ((self.remotefiles[i].path + self.remotefiles[i].name) == (self.currentfile.path + self.currentfile.name)):
                            self.currentfile = copy.copy(self.remotefiles[i])
                    # This second check is necessary, because if the path = /pub2/ and pub2 is a link, it's not found
                    # with the first check !
                    if (self.currentfile.filetype == None) or (self.currentfile.filetype == FTLinkTarget):
                        for i in range(len(self.remotefiles)):
                            if ((self.remotefiles[i].path + self.remotefiles[i].name) + "/" == (self.currentfile.path + self.currentfile.name)):
                                self.currentfile = copy.copy(self.remotefiles[i])
                    # The third check is, if /dir/dir is a directory (no trailing /)
                    if (self.currentfile.filetype == None) or (self.currentfile.filetype == FTLinkTarget):
                        for i in range(len(self.remotefiles)):
                            if ((self.remotefiles[i].path + self.remotefiles[i].name) == (self.currentfile.path + self.currentfile.name + "/")):
                                self.currentfile = copy.copy(self.remotefiles[i])
                else:
                    return result

    # The main mirror loop
    def MainLoop(self):
        while len(self.filelist) > 0:
            self.currentfile = self.filelist[0]
            self.filelist.remove(self.currentfile)
            # Host has changed - open a new connection
            if (self.currentfile.connection != self.connection):
                self.CloseConnection()
                self.connection = self.currentfile.connection
                result = self.OpenConnection()
                if (result == None):
                    self.log4py.debug("[ %s (debug) ] Getting file type of %s%s" % (str(self.title), self.currentfile.path, self.currentfile.name))
                    localresult = self.GetFileType(cache = FALSE, exclude = FALSE, firstURL = TRUE)        # Get the filetype of the first URL
                    while (self.currentfile.filetype == FTLink):
                        self.log4py.info("[ %s ] Following initial link %s%s" % (str(self.title), self.currentfile.path, self.currentfile.name))
                        self.currentfile.name, self.currentfile.path = self.AdoptLinkTarget(self.currentfile.name, self.currentfile.path, self.currentfile.linktarget)
                        self.path = self.currentfile.path + self.currentfile.name
                        self.remotefiles = []
                        self.log4py.debug("[ %s (debug) ] Getting file type of %s%s" % (str(self.title), self.currentfile.path, self.currentfile.name))
                        localresult = self.GetFileType(cache = FALSE, exclude = FALSE)
                    self.CheckDirectory(self.GetLCwd())
            if (result == None):
                if (self.currentfile.connection.protocol == CTFTP):
                    if (self.currentfile.filetype == FTRegularFile):
                        self.DownloadFile()
                    elif (self.currentfile.filetype == FTDirectory):
                        localresult = self.GetRemoteFileList()
                        if (localresult == None):
                            if (len(self.onlylatest) > 0):
                                self.ReduceRemoteFileList()
                            self.CheckDirectory(self.GetLCwd())
                            self.GetLocalFileList()
                            self.DelOldLocalFiles()
                            self.GetRemoteFiles()
                    elif (self.currentfile.filetype == FTLinkTarget):
                        self.GetFileType()
                        if self.currentfile.filetype == None:
                            self.GetFileType()
                        self.CheckDirectory(self.GetLCwd())
                        self.filelist.insert(0, self.currentfile)
                    elif (self.currentfile.filetype == FTLink):
                        self.CreateLink()
                    else:
                        self.log4py.info("[ %s ] Warning: %s%s not found" % (str(self.title), self.currentfile.path, self.currentfile.name))
                        self.AppendLog([LMWarning, self.currentfile.path + self.currentfile.name + " not found"])
                elif (self.currentfile.connection.protocol == CTHTTP):
                    self.log4py.warn("[ %s ] Sorry, but ECLiPt-Mirror doesn't support HTTP (yet) - use wget !" % str(self.title))
        self.CloseConnection()

    # Split a URL
    def SplitURL(self):
        file = File()
        file.log4py.set_loglevel(self.log4py.get_loglevel())

        self.protocol, username, password, self.hostname, self.port, self.path, params, query, fragment = urlparse(self.url)
        
        if (self.username == ""):
            self.username = username
        if (self.password == ""):
            self.password = password

        if (lower(self.protocol) == "ftp") and (self.username == ""):
            self.username = DVUsername
            self.password = DVPassword

        file.path, file.name = os.path.split(self.path)
        if (file.path != "/"):
            file.path = file.path + "/"

        return file

    # Check max delete value
    def ConvertMaxDelete(self, path):
        if self.maxdelete != None:
            if (self.maxdelete[-1:] == "%"):
                self.log4py.info("[ %s ] Calculating disk usage ... " % str(self.title))
                if atoi(self.maxdelete[:-1]) != 0:
                    diskusage = du(path)
                    self.maxdelete = divmod(diskusage, atoi(self.maxdelete[:-1]))[0]
                    self.log4py.info("[ %s ] Disk Usage: %.0f Byte ..." % (str(self.title), diskusage))
                else:
                    self.maxdelete = 0
            elif (lower(self.maxdelete[-1:]) == "k"):
                self.maxdelete = atoi(self.maxdelete[:-1]) * 1024
            elif (lower(self.maxdelete[-1:]) == "m"):
                self.maxdelete = atoi(self.maxdelete[:-1]) * 1024 * 1024
        else:
            self.maxdelete = -1

    # Returns the local directory corresponding to FULL, PATH or NONE
    def GetLCwd(self, hostname = "", path = "", parentdirectory = ""):

        if (hostname == ""):
            hostname = self.currentfile.connection.hostname
        if (path == ""):
            path = self.currentfile.path
        if (parentdirectory == ""):
            parentdirectory = self.directory

        if (self.parentformat == PFFull):
            directory = parentdirectory + hostname + path
        elif (self.parentformat == PFPath):
            directory = parentdirectory + path[1:]
        elif (self.parentformat == PFNone):
            shortpath = path[len(self.path):]
            if (len(shortpath) > 0) and (shortpath[0] == "/"):
                shortpath = shortpath[1:]
            directory = parentdirectory + shortpath

        return directory

    # Start the mirror 
    def Start(self):

        self.log4py.info("[ %s ] Starting mirror" % str(self.title))

        file = self.SplitURL()

        self.StartLog()

        connection = Connection()
        connection.protocol = self.protocol
        connection.hostname = self.hostname
        connection.realhostname = self.realhostname
        connection.port = self.port
        connection.username = self.username
        connection.password = self.password
        connection.status = CSDisconnected

        errorcode = 0
        
        if (connection.hostname == ""):
            self.log4py.error("[ %s ] Invalid URL \"%s\" specified - aborting mirror." % (str(self.title), self.url))
            errorcode = 1

        else:
            file.connection = connection

            self.filelist.append(file)
            self.ConvertMaxDelete(self.GetLCwd(file.connection.hostname, self.path))

            if (self.umask != None):
                umask(self.umask)
                self.log4py.info("[ %s ] Setting umask to %s" % (str(self.title), zfill(oct(self.umask), 3)))

            self.MainLoop()
            
        self.End(errorcode)

    # Stop mirror, write the log and terminate
    def End(self, level):
        self.installer.finish()
        self.WriteLog()

        self.log4py.info("[ %s ] Finishing Mirror" % str(self.title))

        sys.exit(level)

# ----------------------------------------------------------------------
# Definition of the Indexfile-class

class IndexFile:	

    def __init__(self):
        self.directory = None               # Directory where to find the logs
        self.mirrorlogs = []                # List of Logs-files
        self.mirrorstatus = {}              # Status of mirrors
        self.mirrorcategory = {}            # Category of the mirror
        self.categorymirror = {}            # Inverse of mirrorcategory
        self.description = {}               # Description of mirrors
        self.timestamp = {}                 # List of time-stamps for mirrors
        self.logfilename = {}               # Name of the mirror logfile
        self.lastchanged = {}               # Last change of the mirror
        self.log4py = log4py.Logger().get_instance(self)

        self.errorcolor = "#CC0033"         # Default error color
        self.warningcolor = "#009933"       # Default color for warnings

        self.logtemplate = None             # Filename of the log template-file
        self.summarytemplate = None         # Filename of the summary template-file
        self.template = []
        self.templatecategory = []
        self.templateloop = []
        self.templatesummary = []
        self.templatesummaryloop = []

        self.logfileformat = LFPlain        # Format of the indexfile

        self.mailusers = ""                 # Mail info about to changes to some users
        self.mailusercmd = ""               # Command to send mail to users

        self.nrbytesretr = 0L
        self.nrfilesretr = 0L
        self.nrlinkscreated = 0L
        self.nrdirscreated = 0L
        self.nrbytesdel = 0L
        self.nrfilesdel = 0L
        self.nrlinksdel = 0L
        self.nrdirsdel = 0L
        self.totalseconds = 0L
        self.firststarttime = None
        self.lastendtime = None

        self.proxy = None                   # Proxy host and port
        
    # Analyze logfiles to get Title, Status & Category of mirrors
    def AnalyzeLogs(self):
        currentdir = getcwd()
        chdir(self.directory)
        for i in range(len(self.mirrorlogs)):
            datei = open(self.mirrorlogs[i], "r")
            title = ""
            line = "<empty>"
            copyfile = FALSE
            while line != "":
                line = datei.readline()
                stripped = strip(line)
                if (stripped[0:4] == "<!--"):
                    restofline = strip(sub("-->", "", stripped[5:]))
                    if restofline[0:5] == "TITLE":
                        title = restofline[7:]
                        self.logfilename[title] = self.mirrorlogs[i]
                        self.mirrorstatus[title] = MSunknown
                        self.description[title] = "None"
                        self.timestamp[title] = "0"
                    elif restofline[0:6] == "STATUS":
                        self.mirrorstatus[title] = restofline[8:]
                        if restofline[8:][0] in digits:
                            self.lastchanged[title] = restofline[8:]
                            copyfile = TRUE
                    elif restofline[0:8] == "CATEGORY":
                        category = restofline[10:]
                        self.categorymirror[title] = category
                        if self.mirrorcategory.has_key(category):
                            self.mirrorcategory[category].append(title)
                        else:
                            self.mirrorcategory[category] = [title]
                    elif restofline[0:9] == "TIMESTAMP":
                        self.timestamp[title] = restofline[11:]
                    elif restofline[0:11] == "DESCRIPTION":
                        self.description[title] = restofline[13:]
                    elif restofline[0:10] == "STATISTICS":
                        statsline = strip(restofline[12:])
                        splitted = split(statsline)

                        for j in range(len(splitted)):
                            splitted[j] = atol(splitted[j])

                        currentday = localtime(time())
                        midnight = mktime((currentday[0], currentday[1], currentday[2], 0, 0, 0, 0, 0, currentday[8]))

                        if (splitted[8] >= midnight):

                            self.nrbytesretr = self.nrbytesretr + splitted[0]
                            self.nrfilesretr = self.nrfilesretr + splitted[1]
                            self.nrlinkscreated = self.nrlinkscreated + splitted[2]
                            self.nrdirscreated = self.nrdirscreated + splitted[3]
                            self.nrbytesdel = self.nrbytesdel + splitted[4]
                            self.nrfilesdel = self.nrfilesdel + splitted[5]
                            self.nrlinksdel = self.nrlinksdel + splitted[6]
                            self.nrdirsdel = self.nrdirsdel + splitted[7]

                            self.totalseconds = self.totalseconds + (splitted[9] - splitted[8])

                            if (self.firststarttime == None) or (self.firststarttime > splitted[8]):
                                self.firststarttime = splitted[8]
                            if (self.lastendtime == None) or (self.lastendtime < splitted[9]):
                                self.lastendtime = splitted[9]

            datei.close()
            if (copyfile == TRUE): # save last logfile, where changes have occured
                allfiles = listdir(".")
                for j in range(len(allfiles)):
                    if (match(".*" + self.mirrorlogs[i], allfiles[j]) != None):
                        if (allfiles[j][0] in digits):
                            try:
                                unlink(allfiles[j])
                            except:
                                self.log4py.error("[ main ] Index-file: Error while unlinking \"%s\" - Please check the file / directory permissions !" % allfiles[j])
                try:
                    shutil.copy(self.mirrorlogs[i], self.mirrorstatus[title] + "-" + self.mirrorlogs[i])
                except:
                    self.log4py.error("[ main ] Index-file: Error while copying \"%s -> %s\" - Please check the file / directory permissions !" % (self.mirrorlogs[i], self.mirrorstatus[title] + "-" + self.mirrorlogs[i]))
                    
        chdir(currentdir)

    # Read Template file
    def ReadTemplateFile(self):
        file = open(self.logtemplate)
        line = "<empty>"
        self.template = []
        self.templatecategory = []
        self.templateloop = []
        target = self.template
        while (line != ""):
            line = file.readline()
            if (lower(strip(line)) == "$categorystart$"):
                target.append("$categoryloop$")
                target = self.templatecategory
            elif (lower(strip(line)) == "$categoryend$"):
                target = self.template
            elif (lower(strip(line)) == "$mirrorstart$"):
                target.append("$mirrorloop$")
                target = self.templateloop
            elif (lower(strip(line)) == "$mirrorend$"):
                target = self.templatecategory
            elif (find(strip(lower(line)), "<!-- errorcolor:") != -1):
                color = strip(lower(line))
                position = find(color, "<!-- errorcolor:")
                color = color[position + len("<!-- errorcolor:"):]
                position = find(color, "-->")
                color = strip(color[:position])
                self.errorcolor = color
            elif (find(strip(lower(line)), "<!-- warningcolor:") != -1):
                color = strip(lower(line))
                position = find(color, "<!-- warningcolor:")
                color = color[position + len("<!-- warningcolor:"):]
                position = find(color, "-->")
                color = strip(color[:position])
                self.warningcolor = color
            else:
                target.append(line)
        file.close()

    # Read Template file
    def ReadSummaryTemplateFile(self, datei, fileupdated, fileerrors):
        file = open(self.summarytemplate)
        line = "<empty>"
        self.summarytemplate = []
        self.summarytemplateloop = []
        target = self.summarytemplate
        while (line != ""):
            line = file.readline()
            line = sub("\$indexfilename\$", datei.name, line)
            line = sub("\$updatedfilename\$", fileupdated.name, line)
            line = sub("\$errorfilename\$", fileerrors.name, line)
            if (lower(strip(line)) == "$mirrorstart$"):
                target.append("$mirrorloop$")
                target = self.summarytemplateloop
            elif (lower(strip(line)) == "$mirrorend$"):
                target = self.summarytemplate
            else:
                target.append(line)
        file.close()

    # Create the master index-file
    def CreateMasterIndexFile(self, categorylist, duration, downloaded, deleted, downloadrate, datei, categorykeys, fileupdated, fileerrors):
        for i in range(len(self.template)):
            line = self.template[i]
            if (line == "$categoryloop$"):                           # Add lines for each category
                for j in range(len(categorykeys)):
                    category = categorykeys[j]
                    mirrorlist = self.mirrorcategory[categorykeys[j]]
                    mirrorlist.sort()
                    for k in range(len(self.templatecategory)):
                        line = self.templatecategory[k]
                        if (line == "$mirrorloop$"):                 # Add lines for each mirror
                            for l in range(len(mirrorlist)):
                                title = mirrorlist[l]
                                update = "Unknown"
                                if (self.logfileformat == LFPlain):
                                    mirrortitle = title
                                    if self.mirrorstatus[title] == MSunknown: status = "in progress"
                                    elif self.mirrorstatus[title] == MSnotloggedin: status = "not mirrored"
                                    elif self.mirrorstatus[title] == MSunchanged: status = "unchanged"
                                    else: status = "updated"
                                    if self.lastchanged.has_key(title):
                                        update = ctime(atof(self.lastchanged[title]))
                                else:
                                    mirrortitle = "<a href=\"%s\">%s</a>" % (self.logfilename[title], title)
                                    if self.mirrorstatus[title] == MSunknown: status = "<font color=\"%s\">in progress</font>" % self.warningcolor
                                    elif self.mirrorstatus[title] == MSnotloggedin: status = "<font color=\"%s\">not mirrored</font>" % self.errorcolor
                                    elif self.mirrorstatus[title] == MSunchanged: status = "unchanged"
                                    else: status = "<font color=\"%s\">updated</font>" % self.warningcolor
                                    if self.lastchanged.has_key(title):
                                        update = "<a href=\"%s\">%s</a>" % (self.lastchanged[title] + "-" + self.logfilename[title], ctime(atof(self.lastchanged[title])))
                                for m in range(len(self.templateloop)):
                                    line = self.templateloop[m]
                                    line = sub("\$status\$", status, line)
                                    line = sub("\$title\$", mirrortitle, line)
                                    line = sub("\$update\$", update, line)
                                    datei.write(line)
                        else:
                            line = sub("\$category\$", category, line)
                            datei.write(line)
            else:
                line = sub("\$categorylist\$", categorylist, line)
                line = sub("\$lastupdate\$", ctime(time()), line)
                if (self.firststarttime != None):
		    line = sub("\$starttime\$", ctime(self.firststarttime), line)
		else:
		    line = sub("\$starttime\$", "Unknown", line)
		if (self.lastendtime != None):
		    line = sub("\$endtime\$", ctime(self.lastendtime), line)
		else:
                    line = sub("\$endtime\$", "Unknown", line)

                line = sub("\$duration\$", duration, line)
                line = sub("\$downloadstats\$", downloaded, line)
                line = sub("\$deletestats\$", deleted, line)
                line = sub("\$downloadrate\$", downloadrate, line)
                line = sub("\$proxy\$", str(self.proxy), line)
                line = sub("\$indexfilename\$", datei.name, line)
                line = sub("\$updatedfilename\$", fileupdated.name, line)
                line = sub("\$errorfilename\$", fileerrors.name, line)
                datei.write(line)

    # Create errors & updated file
    def CreateSummaryFile(self, fileupdated, fileerrors):

        updatedlist = {}
        errorlist = {}
        for i in range(len(self.mirrorcategory.keys())):
            category = self.mirrorcategory.keys()[i]
            for j in range(len(self.mirrorcategory[category])):
                title = self.mirrorcategory[category][j]
                timestamp = self.timestamp[title]
                description = self.description[title]
                if self.mirrorstatus[title] == MSnotloggedin:
                    errorlist[timestamp] = title
                elif (self.mirrorstatus[title] != MSunchanged) and (self.mirrorstatus[title] != MSunknown):
                    updatedlist[timestamp] = title
                
        for i in range(len(self.summarytemplate)):
            line = self.summarytemplate[i]
            if (line == "$mirrorloop$"):
                
                # Write updated file information

                if len(updatedlist.keys()) == 0:
                    updatedlist["0"] = "None"
                    self.categorymirror["None"] = "None"
                    self.logfilename["None"] = ""
                    self.description["None"] = "No packages have been updated"
                    
                timelist = updatedlist.keys()
                timelist.sort()

                for j in range(len(timelist)):
                    timestamp = timelist[j]
                    title = updatedlist[timestamp]
                    mirrortitle = title
                    categorylink = self.categorymirror[title]
                    if (self.logfileformat == LFHTML) or (self.logfileformat == LFPHP):
                        mirrortitle = "<a href=\"%s\">%s</a>" % (self.logfilename[title], title)
                        if (self.logfileformat == LFHTML):
                            categorylink = "<a href=\"%s\">%s</a>" % ("index" + LEHTML, self.categorymirror[title])
                        elif (self.logfileformat == LFPHP):
                            categorylink = "<a href=\"%s#%s\">%s</a>" % ("index" + LEPHP, self.categorymirror[title], self.categorymirror[title])
                            
                    for k in range(len(self.summarytemplateloop)):
                        line = self.summarytemplateloop[k]
                        line = sub("\$update\$", asctime(localtime(atof(timestamp))), line)
                        line = sub("\$title\$", mirrortitle, line)
                        line = sub("\$category\$", categorylink, line)
                        line = sub("\$description\$", self.description[title], line)
                        fileupdated.write(line)

                # Write error file information

                if len(errorlist.keys()) == 0:
                    errorlist["0"] = "None"
                    self.categorymirror["None"] = "None"
                    self.logfilename["None"] = ""
                    self.description["None"] = "No errors have occured"

                timelist = errorlist.keys()
                timelist.sort()
                for j in range(len(timelist)):
                    timestamp = timelist[j]
                    title = errorlist[timestamp]
                    mirrortitle = title
                    categorylink = self.categorymirror[title]
                    if (self.logfileformat == LFHTML) or (self.logfileformat == LFPHP):
                        mirrortitle = "<a href=\"%s\">%s</a>" % (self.logfilename[title], title)
                        if (self.logfileformat == LFHTML):
                            categorylink = "<a href=\"%s\">%s</a>" % ("index" + LEHTML, self.categorymirror[title])
                        elif (self.logfileformat == LFPHP):
                            categorylink = "<a href=\"%s#%s\">%s</a>" % ("index" + LEPHP, self.categorymirror[title], self.categorymirror[title])

                    for k in range(len(self.summarytemplateloop)):
                        line = self.summarytemplateloop[k]
                        line = sub("\$update\$", asctime(localtime(atof(timestamp))), line)
                        line = sub("\$title\$", mirrortitle, line)
                        line = sub("\$category\$", categorylink, line)
                        line = sub("\$description\$", self.description[title], line)
                        fileerrors.write(line)

            else:
                line = sub("\$lastupdate\$", ctime(time()), line)
                if find(line, "$summary$") != -1:
                    lineupdated = sub("\$summary\$", "Updated mirrors", line)
                    lineerrors = sub("\$summary\$", "Failed mirrors", line)
                    fileupdated.write(lineupdated)
                    fileerrors.write(lineerrors)
                else:
                    fileupdated.write(line)
                    fileerrors.write(line)

    # Create the new Index-file
    def CreateFile(self):
        currentdir = getcwd()
        chdir(self.directory)

        if (self.logtemplate == None) or (self.summarytemplate == None):
            self.log4py.error("[ Index ] Logfile: Error: No template file specified !")
            return

        try:
            if (self.logfileformat == LFHTML):
                datei = open("index" + LEHTML, "w")
                fileupdated = open("updated" + LEHTML, "w")
                fileerrors = open("errors" + LEHTML, "w")
            elif (self.logfileformat == LFPHP):
                datei = open("index" + LEPHP, "w")
                fileupdated = open("updated" + LEPHP, "w")
                fileerrors = open("errors" + LEPHP, "w")
            elif (self.logfileformat == LFPlain):
                datei = open("index" + LEPlain, "w")
                fileupdated = open("updated" + LEPlain, "w")
                fileerrors = open("errors" + LEPlain, "w")
        except:
            self.log4py.error("[ Index ] Logfile: Couldn't open index file !")
            return

        self.ReadTemplateFile()
        self.ReadSummaryTemplateFile(datei, fileupdated, fileerrors)

        # Calculate the time
        minutes, seconds = divmod(self.totalseconds, 60)
        hours, minutes = divmod(minutes, 60)

        # Create Format specific text
        if (self.logfileformat == LFPlain):
            duration = "%2d Hour(s), %2d Minute(s) and %2d Second(s)" % (hours, minutes, seconds)
            downloaded = "%s Byte(s) in %.0f File(s), %.0f Link(s) and %.0f Directories ..." % (lng2str(self.nrbytesretr), self.nrfilesretr, self.nrlinkscreated, self.nrdirscreated)
            deleted = "%s Byte(s) in %.0f File(s), %.0f Link(s) and %.0f Directories ..." % (lng2str(self.nrbytesdel), self.nrfilesdel, self.nrlinksdel, self.nrdirsdel)
        else:
            duration = "<b>%2d</b> Hour(s), <b>%2d</b> Minute(s) and <b>%2d</b> Second(s)" % (hours, minutes, seconds)
            downloaded = "<b>%s</b> Byte(s) in <b>%.0f</b> File(s), <b>%.0f</b> Link(s) and <b>%.0f</b> Directories ..." % (lng2str(self.nrbytesretr), self.nrfilesretr, self.nrlinkscreated, self.nrdirscreated)
            deleted = "<b>%s</b> Byte(s) in <b>%.0f</b> File(s), <b>%.0f</b> Link(s) and <b>%.0f</b> Directories ..." % (lng2str(self.nrbytesdel), self.nrfilesdel, self.nrlinksdel, self.nrdirsdel)

        # Calculate download rate
        if (self.nrbytesretr > 0):
            if (self.totalseconds > 0):
                rate = self.nrbytesretr / self.totalseconds
            else:
                rate = 0
            kbytepersec = int(rate / 1024)
            if (kbytepersec > 0):
                if (self.logfileformat == LFPlain):
                    downloadrate = "%.0f KByte/sec" % int(kbytepersec)
                else:
                    downloadrate = "<b>%.0f</b> KByte/sec" % int(kbytepersec)
            else:
                if (self.logfileformat == LFPlain):
                    downloadrate = "%.0f Byte/sec" % int(rate)
                else:
                    downloadrate = "<b>%.0f</b> Byte/sec" % int(rate)
        else:
            downloadrate = "Nothing downloaded"

        categorykeys = self.mirrorcategory.keys()
        categorykeys.sort()

        categorylist = ""

        for i in range(len(categorykeys)):
            if (self.logfileformat == LFPlain):
                categorylist = "%s%s" % (categorylist, categorykeys[i])
            else:
                categorylist = "%s<a href=\"#%s\">%s</a>" % (categorylist, categorykeys[i], categorykeys[i])
            if (i < (len(categorykeys) - 1)):
                categorylist = "%s | " % categorylist

        self.CreateMasterIndexFile(categorylist, duration, downloaded, deleted, downloadrate, datei, categorykeys, fileupdated, fileerrors)
        self.CreateSummaryFile(fileupdated, fileerrors)

        datei.close()
        fileupdated.close()
        fileerrors.close()

        chdir(currentdir)

    def GetLastlog(self):
        currentdir = getcwd()
        chdir(self.directory)
        if os.path.exists(LastlogFilename):
            datei = open(LastlogFilename, "r")
            line = "<empty>"
            while line != "":
                line = datei.readline()
                if strip(line) != "":
                    splitted = split(strip(line), ";")
                    if not self.lastchanged.has_key(splitted[0]):
                        self.lastchanged[splitted[0]] = splitted[1]
            datei.close()
        chdir(currentdir)

    def WriteLastlog(self):
        currentdir = getcwd()
        chdir(self.directory)
        try:
            datei = open(LastlogFilename, "w")
        except:
            self.log4py.error("[ Index ] Logfile: Couldn't write lastlog file !")
            return
        alltitles = self.lastchanged.keys()
        for i in range(len(alltitles)):
            datei.write(alltitles[i] + ";" + self.lastchanged[alltitles[i]] + "\n")
        datei.close()
        chdir(currentdir)

    # Send mail to users, if things have changed or if emirror couldn't log in
    def MailUsers(self):

        changed = []
        notloggedin = []

        for i in range(len(self.mirrorstatus.keys())):
            key = self.mirrorstatus.keys()[i]
            if (self.mirrorstatus[key] == MSnotloggedin):
                notloggedin.append(key)
            elif ((self.mirrorstatus[key] != MSunknown) and (self.mirrorstatus[key] != MSunchanged)):
                changed.append(key)

        changed.sort()
        notloggedin.sort()

        if (len(changed) > 0) or (len(notloggedin) > 0):
            self.log4py.info("[ Main ] Mirrors have changed / updated - sending mail.")

            parser = ConfigParser()
            parser.read(self.mailusers)
            emailaddresses = split(parser.get("email", "list"), ",")

            filename = mktemp()
            file = open(filename, "w")
            file.write("\nWelcome to the EMirror EMail-Update Service !\n\n")

            # Calculate the time
            minutes, seconds = divmod(self.totalseconds, 60)
            hours, minutes = divmod(minutes, 60)

            # Create Format specific text
            duration = "%2d Hour(s), %2d Minute(s) and %2d Second(s)" % (hours, minutes, seconds)
            downloaded = "%s Byte(s) in %.0f File(s), %.0f Link(s) and %.0f Directories ..." % (lng2str(self.nrbytesretr), self.nrfilesretr, self.nrlinkscreated, self.nrdirscreated)
            deleted = "%s Byte(s) in %.0f File(s), %.0f Link(s) and %.0f Directories ..." % (lng2str(self.nrbytesdel), self.nrfilesdel, self.nrlinksdel, self.nrdirsdel)

            if (self.nrbytesretr > 0):
                if (self.totalseconds > 0):
                    rate = self.nrbytesretr / self.totalseconds
                else:
                    rate = 0
                kbytepersec = int(rate / 1024)
                if (kbytepersec > 0):
                    downloadrate = "%.0f KByte/sec" % int(kbytepersec)
                else:
                    downloadrate = "%.0f Byte/sec" % int(rate)
            else:
                downloadrate = "Nothing downloaded"
    
            if (self.firststarttime != None):
                file.write("Start of first mirror: %s\n" % ctime(self.firststarttime))
            else:
                file.write("Start of first mirror: <not set>\n")
            if (self.lastendtime != None):
                file.write("   End of last mirror: %s\n" % ctime(self.lastendtime))
            else:
                file.write("   End of last mirror: <not set>\n")
            file.write("       Total duration: %s\n" % duration)
            file.write("           Downloaded: %s\n" % downloaded)
            file.write("   Avg. download rate: %s\n" % downloadrate)
            file.write("              Deleted: %s\n" % deleted)
            file.write("      Proxy host/port: %s\n\n" % str(self.proxy))

            file.write("The following list of mirrors has changed:\n\n")
            for i in range(len(changed)):
                file.write("    %s [%s]\n" % (changed[i], self.categorymirror[changed[i]]))
            if (len(changed) == 0):
                file.write("    [nothing has changed]\n")
            file.write("\nEMirror couldn't log in into those mirrors. This probably doesn't mean\nanything, but if it happens more often, you maybe have to check that !\n\n")
            for i in range(len(notloggedin)):
                file.write("    %s [%s]\n" % (notloggedin[i], self.categorymirror[notloggedin[i]]))
            if (len(notloggedin) == 0):
                file.write("    [none]\n")
            file.close()
            for i in range(len(emailaddresses)):
                emailaddress = emailaddresses[i]
                self.log4py.debug("[ Main (debug) ] Executing mail-command: cat %s | %s %s " % (filename, self.mailusercmd, emailaddress))
                system("cat %s | %s %s" % (filename, self.mailusercmd, emailaddress))

            unlink(filename)

    def Start(self):
        if (self.directory[-1:] != "/"):
            self.directory = self.directory + "/"
        if not os.path.exists(self.directory):
            self.log4py.error("[ Main ] Error: Directory \"%s\" doesn't exist !" % self.directory)
            exit(1)

        allfiles = listdir(self.directory)

        excludefiles = compile("index.*|.*jpg$|.*gif$")
        for i in range(len(allfiles)):
            if (allfiles[i][0] not in digits) and (not excludefiles.match(allfiles[i])):
                self.mirrorlogs.append(allfiles[i])

        self.AnalyzeLogs()
        self.GetLastlog()
        self.CreateFile()
        self.WriteLastlog()
        if (self.mailusers != "") and (self.mailusercmd != ""):
            self.MailUsers()
        else:
            if (self.mailusers != "") or (self.mailusercmd != ""):
                self.log4py.warn("[ Main ] Warning: you have to set mail-users AND mail-user-cmd for sending mail !")

# ----------------------------------------------------------------------
# Definition of the Program-Class:

# A global Variable for the PIDs of concurrent children
children = []

class Program:

    # Initialisation of variables
    def __init__(self):
        self.mirror = Mirror()                   # The mirror to set parameters
        self.mirrorlist = []                     # List of mirrors
        self.indexfile = IndexFile()             # IndexFile to create
        self.configfile = None                   # Master configuration file
        self.configdirectory = None              # Directory containing configuration files
        self.maxfork = 1                         # How many parallel mirrors shall I run
        self.uid = None                          # The UID of the user who executes the program
        self.log4py = log4py.Logger().get_instance(self)
        self.pidfile = "%s/.emirror.pid" % environ["HOME"]    # Filename of the PID-File
        self.killprocesses = TRUE                # Kill old processes on startup or not

    # Print some information about the program & syntax
    def PrintInfo(self, message, errormsg = None):
        sys.stdout.write("%s\n" % PIHeader)
        if (errormsg != None):
            sys.stdout.write("%s. Try emirror --help for a list of possible parameters.\n" % capitalize(errormsg))
        else:
            sys.stdout.write("%s\n" % message)
            sys.stdout.write("%s\n" % PIOptions)
        exit(0)

    # checks wether a file exists and reports an error if not
    def FileExists(self, file):
        if (not os.path.exists(file)):
            self.log4py.warn("[ Main ] Warning: file \"%s\" does not exist." % file)
            return ""
        else:
            if (find(file, "/") != -1):
                return file
            else:
                return getcwd() + "/" + file

    # Check command line / configuration file options
    def CheckOptions(self, parameter, value, filename = "", option = None):
        if (parameter[0:2] == "--"):
            parameter = parameter[2:]

        # Startup
        if (parameter == "-V" or parameter == "version"):
            sys.stdout.write("%s\n" % PIHeader)
            sys.stdout.write("%s\n" % PIRevLog)
            exit(0)
        elif (parameter == "-h" or parameter == "help"):
            self.PrintInfo(PIUsage)
        elif (parameter == "-v" or parameter == "verbose"):
            self.log4py.set_loglevel(log4py.LOGLEVEL_VERBOSE)
            self.mirror.log4py.set_loglevel(log4py.LOGLEVEL_VERBOSE)
            self.mirror.installer.log4py.set_loglevel(log4py.LOGLEVEL_VERBOSE)
            self.indexfile.log4py.set_loglevel(log4py.LOGLEVEL_VERBOSE)

        # Logging and input files
        elif (parameter == "-o" or parameter == "output-file"):
            self.mirror.logfilename = value
        elif (parameter == "-O" or parameter == "output-directory"):
            if value[-1:] != "/":
                value = value + "/"
            self.mirror.logdirectory = value
        elif (parameter == "-f" or parameter == "output-format"):
            format = lower(value)
            if (format != LFHTML) and (format != LFPlain) and (format != LFPHP):
                self.log4py.warn("[ Main ] Warning: invalid logfile format, choose HTML or Plain.")
                format = LFHTML
            self.mirror.logfileformat = format
            self.indexfile.logfileformat = format
        elif (parameter == "-c" or parameter == "config-file"):
            self.configfile = value
        elif (parameter == "-d" or parameter == "config-directory"):
            self.configdirectory = value
        elif (parameter == "-n" or parameter == "index-directory"):
            if value[-1:] != "/":
                value = value + "/"
            self.indexfile.directory = value
        elif (parameter == "category"):
            self.mirror.category = value
        elif (parameter == "description"):
            self.mirror.description = value
        elif (parameter == "title"):
            self.mirror.title = value
            self.mirror.installer.title = value
        elif (parameter == "email"):
            self.mirror.adminemail = value
        elif (parameter == "mail-users"):
            self.indexfile.mailusers = self.FileExists(value)
        elif (parameter == "mail-user-cmd"):
            self.indexfile.mailusercmd = value
        elif (parameter == "log-template"):
            self.mirror.logtemplate = self.FileExists(value)
        elif (parameter == "index-template"):
            self.indexfile.logtemplate = self.FileExists(value)
        elif (parameter == "summary-template"):
            self.indexfile.summarytemplate = self.FileExists(value)

        # Download and FTP options
        elif (parameter == "-P" or parameter == "parent"):
            if value[-1:] != "/":
                value = value + "/"
            if value[0] != "/":
                self.log4py.warn("[ Main ] Warning: use absolut values for the \"parent\" parameter !.")
                value = "%s/%s" % (os.getcwd(), value)
            self.mirror.directory = value
        elif (parameter == "-D" or parameter == "download-url"):
            if value[-1:] != "/":
                value = value + "/"
            self.mirror.localurl = value
        elif (parameter == "-t" or parameter == "tries"):
            self.mirror.retries = atoi(value)
        elif (parameter == "-w" or parameter == "wait"):
            self.mirror.delay = atoi(value)
        elif (parameter == "-U" or parameter == "umask"):
            self.mirror.umask = atoi(value, 8)
        elif (parameter == "-r" or parameter == "recursive"):
            if (str2bool(value) == TRUE) or (value == ""):
                self.mirror.recursive = TRUE
            elif (str2bool(value) == FALSE):
                self.mirror.recursive = FALSE
            else:
                self.log4py.warn("[ Main ] Warning: invalid value for -r/--recursive. Use Yes/No or TRUE/False.")

        elif (parameter == "-e" or parameter == "exclude"):
            try:
                self.mirror.exclude = compile(unfold(value))
            except:
                self.log4py.error("[ Main ] Error in exclude-regexp: %s" % value)
                exit(1)

        elif (parameter == "-i" or parameter == "include"):
            try:
                self.mirror.include = compile(unfold(value))
            except:
                self.log4py.error("[ Main ] Error in include-regexp: %s" % value)
                exit(1)

        elif (parameter == "ignore"):
            try:
                self.mirror.ignore = compile(unfold(value))
            except:
                self.log4py.error("[ Main ] Error in ignore-regexp: %s" % value)
                exit(1)

        elif (parameter == "proxy"):
            splitted = split(value, ":")
            if (len(splitted) != 2):
                self.log4py.warn("[ Main ] Invalid value for \"Proxy\", use <hostname:port> !")
            else:
                self.mirror.proxy = value
                self.indexfile.proxy = value
                if self.mirror.passivemode == FALSE:
                    self.mirror.passivemode = TRUE
                    
        elif (parameter == "test"):
            if (str2bool(value) == TRUE) or (value == ""):
                self.mirror.test = TRUE
            elif (str2bool(value) == FALSE):
                self.mirror.test = FALSE
            else:
                self.log4py.warn("[ Main ] Warning: invalid value for --test. Use Yes/No or True/False.")

        # Set of only-latest values
        elif (parameter == "-l" or parameter == "only-latest"):
            try:
                regexp = compile(unfold(value))
                self.mirror.onlylatest.append(regexp)
            except:
                self.log4py.error("[ Main ] Error in only-latest-regexp: %s" % value)
                exit(1)

        # Set of auto install values
        elif (parameter == "auto-install"):
            self.mirror.installer.add(option, value)

        elif (parameter == "-u" or parameter == "username"):
            self.mirror.username = value
        elif (parameter == "-p" or parameter == "password"):
            self.mirror.password = value
        elif (parameter == "-m" or parameter == "max-parallel"):
            self.maxfork = atoi(value)
        elif (parameter == "-R" or parameter == "real-hostname"):
            self.mirror.realhostname = value
        elif (parameter == "-M" or parameter == "max-days"):
            self.mirror.maxdays = atoi(value)
        elif (parameter == "-A" or parameter == "max-age"):
            self.mirror.maxage = atoi(value)
        elif (parameter == "-E" or parameter == "max-delete"):
            if (value[-1:] != "%") and (lower(value[-1:]) != "k") and (lower(value[-1:]) != "m"):
                self.log4py.warn("[ Main ] Warning: invalid value for --max-delete (has to end in M, K or %).")
            else:
                self.mirror.maxdelete = value
        elif (parameter == "-F" or parameter == "follow-symlinks"):
            if (str2bool(value) == TRUE) or (value == ""):
                self.mirror.followsymlinks = TRUE
            elif (str2bool(value) == FALSE):
                self.mirror.followsymlinks = FALSE
            else:
                self.log4py.warn("[ Main ] Warning: invalid value for -F/--follow-symlinks. Use Yes/No or True/False.")
        elif (parameter == "-s" or parameter == "parent-format"):
            if (lower(value) == "full") or (lower(value) == "path") or (lower(value) == "none"):
                self.mirror.parentformat = lower(value)
            else:
                self.log4py.warn("[ Main ] Warning: invalid value for -s/--parent-format: use FULL, PATH or NONE.")
        elif (parameter == "ignore-mdtm"):
            if (str2bool(value) == TRUE) or (value == ""):
                self.mirror.ignoremdtm = TRUE
            elif (str2bool(value) == FALSE):
                self.mirror.ignoremdtm = FALSE
            else:
                self.log4py.warn("[ Main ] Warning: invalid value for --ignore-mdtm: use Yes/No or True/False.")
        elif (parameter == "owner"):
            if (self.uid != 0):
                self.log4py.warn("[ Main ] Warning: only user \"root\" can set the owner of files.")
            else:
                if (find(value, ":") == -1):
                    self.log4py.warn("[ Main ] Warning: wrong format for owner: use UID:GID")
                else:
                    splitted = split(value, ":")
                    self.mirror.owner = atoi(splitted[0])
                    self.mirror.group = atoi(splitted[1])

        elif (parameter == "delete-local"):
            if (str2bool(value) == TRUE) or (value == ""):
                self.mirror.deletelocal = TRUE
            elif (str2bool(value) == FALSE):
                self.mirror.deletelocal = FALSE
            else:
                self.log4py.warn("[ Main ] Warning: invalid value for --delete-local: use Yes/No or True/False.")
        elif (parameter == "passive"):
            if (str2bool(value) == TRUE) or (value == ""):
                self.mirror.passivemode = TRUE
            elif (str2bool(value) == FALSE):
                self.mirror.passivemode = FALSE
            else:
                self.log4py.warn("[ Main ] Warning: invalid value for --passive: use Yes/No or True/False.")
        elif (parameter == "continue"):
            if (str2bool(value) == TRUE) or (value == ""):
                self.mirror.continueftp = TRUE
            elif (str2bool(value) == FALSE):
                self.mirror.continueftp = FALSE
            else:
                self.log4py.warn("[ Main ] Warning: invalid value for --continue: use Yes/No or True/False.")
                
        # Error handling options
        elif (parameter == "error-mailcmd"):
            self.mirror.errormailcmd = value

        # Info handling
        elif (parameter == "info-mailcmd"):
            self.mirror.infomailcmd = value
            
        # Debug options
        elif (parameter == "debug"):
            self.log4py.set_loglevel(log4py.LOGLEVEL_DEBUG)
            self.mirror.log4py.set_loglevel(log4py.LOGLEVEL_DEBUG)
            self.mirror.installer.log4py.set_loglevel(log4py.LOGLEVEL_DEBUG)
            self.indexfile.log4py.set_loglevel(log4py.LOGLEVEL_DEBUG)

        elif (parameter == "noexception"):
            self.mirror.noexception = TRUE

        # Other options
        elif (parameter == "url"):
            self.mirror.url = value
        elif (parameter == "-k" or parameter == "no-kill"):
            self.killprocesses = FALSE
        elif (parameter == "pid-file"):
            self.pidfile = value

        else:
            log("%s" % PIHeader)
            if (filename != ""):
                self.log4py.error("[ Main ] Error in config-file %s - invalid parameter \"%s\"\n" % (filename, parameter))
            else:
                self.log4py.error("[ Main ] Error - invalid parameter \"%s\"\n\n" % parameter)
            exit(1)

    # Parse the command line, if a url has been specified
    def ParseCmdLine(self):
        if environ.has_key(EMirrorEnvOptions):
            envargv = split(environ[EMirrorEnvOptions])
            fullargv = envargv + argv[1:]
        else:
            fullargv = argv[1:]
        try:
            optlist, args = getopt.getopt(fullargv,
                                          'Vhvo:O:f:c:d:n:P:D:t:w:re:i:lu:p:m:R:M:A:E:Fs:U:k',
                                          ['version', 'help', 'verbose', 'output-file=', 'output-directory=', 'output-format=', 
                                           'config-file=', 'config-directory=', 'index-directory=', 'category=',
                                           'description=', 'title=', 'email=', 'parent=', 'download-url=', 'tries=', 'wait=', 
                                           'recursive', 'exclude=', 'include=', 'only-latest', 'username=', 'password=',
                                           'max-parallel=', 'real-hostname=',
                                           'max-days=', 'max-age=', 'passive', 'umask=',
                                           'max-delete=', 'log-template=', 'index-template=',
                                           'follow-symlinks=', 'parent-format=', 'error-mailcmd=', 'ignore-mdtm', 'owner=', 
                                           'delete-local', 'no-kill', 'pid-file=',
                                           'mail-users=', 'mail-user-cmd=', 'debug', 'noexception', 'proxy=', 'continue', 
                                           'info-mailcmd=', 'test', 'ignore='])
        except getopt.error, detail:
            self.PrintInfo(PIUsage, detail)

        for flag, arg in optlist:
            self.CheckOptions(flag, arg)

        if (self.configfile == None) and (self.indexfile.directory == None):
            if (len(args) != 1):
                self.PrintInfo(PIUsage)
        if (len(args) > 0):
            self.mirror.url = args[0]
            if (lower(self.mirror.url[0:6]) != "ftp://") and (lower(self.mirror.url[0:7]) != "http://"):
                self.log4py.error("%s" % PIHeader)
                self.log4py.error("Error: specified protocol not supported ...\n")
                exit(1)
            self.mirrorlist.append(self.mirror)

    # Read a configuration file
    def ReadConfigFile(self, configfilename, general = None):
        found = FALSE
        if os.path.exists(configfilename):
            found = TRUE
        if (found == FALSE):
            if environ.has_key(EMirrorEnvironment):
                emirrorcfgpath = environ[EMirrorEnvironment]
                if emirrorcfgpath[-1:] != "/":
                    emirrorcfgpath = emirrorcfgpath + "/"
                configfilename = emirrorcfgpath + configfilename
            if not os.path.exists(configfilename):
                self.log4py.error("%s" % PIHeader)
                self.log4py.error("Configuration file \"%s\" does't exist ...\n" % configfilename)
                exit(1)

        if (general == None):
            parser = ConfigParser()
        else:
            parser = general
        try:
            parser.read(configfilename)
        except:
            self.log4py.error("[ Main ] Error: invalid config-file: %s" % configfilename)
            exit(1)

        for i in range(len(parser.sections())):
            section = parser.sections()[i]
            for j in range(len(parser.options(section))):
                option = parser.options(section)[j]
                if (option != "name") and (option != "__name__"):               # name is the name of the section
                    if (lower(section) == "only-latest"):
                        self.CheckOptions("only-latest", parser.get(section, option), configfilename)
                    elif (lower(section) == "auto-install"):
                        self.CheckOptions("auto-install", parser.get(section, option), configfilename, option)
                    else:
                        self.CheckOptions(option, parser.get(section, option), configfilename)

    # Get a list of configuration files ...
    def GetConfigFiles(self, defaults):

        if not os.path.exists(self.configdirectory):
            self.log4py.error(PIHeader)
            self.log4py.error("\n[ Main ] Error: Directory / File %s doesn't exist !\n" % self.configdirectory)
            exit(1)

        if os.path.isdir(self.configdirectory):
            if (self.configdirectory[-1:] != "/"):
                self.configdirectory = "%s/" % self.configdirectory
            files = listdir(self.configdirectory, incdirs = FALSE, inclinks = FALSE)
        else:
            files = [os.path.basename(self.configdirectory)]
            self.configdirectory = "%s/" % os.path.dirname(self.configdirectory)
            
        defaultmirror = copy.deepcopy(self.mirror)
        count = 0
        for i in range(len(files)):
            filename = files[i]
            if (filename[-1:] != "~"):
                self.mirror = copy.deepcopy(defaultmirror)
                self.ReadConfigFile(self.configdirectory + filename, defaults)
                self.mirrorlist.append(self.mirror)
                count = count + 1
        self.log4py.debug("[ Main (debug) ] %.0f configuration file(s) read" % count)

    # Signal Handler for SIGCHLD 
    def SigChldHandler(self, signal_number, stack_frame):
        # Reinstall the sigchild handler, cuz it gets deleted on libc5/sysv machines
        signal.signal(signal.SIGCHLD, self.SigChldHandler)
        global children
        pid = -1
        while (pid != 0):
            try:
                pid = os.waitpid(0, os.WNOHANG)[0]
            except:
                pid = 0
            if (pid != 0):
                if pid in children:
                    self.log4py.debug("[ Main (debug) ] Child finished")
                    children.remove(pid)

    # Delete the PID file
    def DeletePIDFile(self):
        if os.path.exists(self.pidfile):
            remove(self.pidfile)
            self.log4py.debug("[ Main (debug) ] PID File deleted")

    # Define a Handler for SigTerm for the parent
    def SigTermHandler(self, signal_number, stack_frame):
        self.log4py.info("[ Main ] main program aborted or killed - terminating children")
        global children, forkchildren
        self.log4py.debug("[ Main (debug) ] children's PIDs: %s" % children)
        forkchildren = FALSE
        for i in range(len(children)):                                          # Kill Children
            try:
                kill(children[i], signal.SIGTERM)
                kill(children[i], signal.SIGKILL)                               # Kill them again, just in case the child survived ;-)
            except:
                self.log4py.debug("[ Main (debug) ]: kill -9 %s (child) failed" % children[i])
                pass
        self.DeletePIDFile()
        exit(1)

    # Define a Handler for SigTerm for the children
    def SigTermHandlerChild(self, signal_number, stack_frame):
        exit(1)

    # Kills old EMirror Master Processes
    def KillOldProcesses(self):
        self.log4py.debug("[ Main (debug) ] reading process table & killing old mirror-processes")
        processes = process.ProcessTable()
        processes.read()
        processes.filter(filters = {"CMD": ["python", "emirror"]}, filtermode = process.FILTERMODESUB)
        ppids = processes.ppids().keys()
        for i in range(len(ppids)):
            ppid = atoi(ppids[i])
            if (ppid in processes.pids()):
                self.log4py.info("[ Main ] killing old EMirror master-process with pid %d" % ppid)
                kill(ppid, signal.SIGTERM)

    # Start mirroring ;-)
    def Start(self):
        sys.stderr = sys.stdout

        self.uid = getuid()
        self.ParseCmdLine()

        if self.configfile != None:
            self.log4py.debug("[ Main (debug) ] reading main configuration file")
            defaults = self.ReadConfigFile(self.configfile)

        if (self.configdirectory != None):
            self.log4py.debug("[ Main (debug) ] reading other configuration files")
            self.GetConfigFiles(defaults)

        if (self.maxfork == 1):
            self.log4py.info("[ Main ] ECLiPt Mirroring Tool Version %s - starting 1 copy" % PIVersion)
        else:
            self.log4py.info("[ Main ] ECLiPt Mirroring Tool Version %s - starting max. %d copies" % (PIVersion, self.maxfork))

        if (self.killprocesses == TRUE):
            self.KillOldProcesses()

        signal.signal(signal.SIGINT, self.SigTermHandler)
        signal.signal(signal.SIGTERM, self.SigTermHandler)

        signal.signal(signal.SIGCHLD, self.SigChldHandler)

        self.log4py.debug("[ Main (debug) ] writing pid-file (my pid is %d)" % getpid())
        runfile = open(self.pidfile, "w")
        runfile.write("%d" % getpid())
        runfile.close()

        global children, forkchildren
        forkchildren = TRUE
        children = []

        while len(self.mirrorlist) > 0:
            if (len(children) < self.maxfork) and (forkchildren == TRUE):
                self.log4py.debug("[ Main (debug) ] starting new child")
                mirror = self.mirrorlist[0]
                self.mirrorlist = self.mirrorlist[1:]
                self.log4py.debug("[ Main (debug) ] %d mirrors remaining" % len(self.mirrorlist))
                result = os.fork()
                if (result == 0):
                    signal.signal(signal.SIGINT, self.SigTermHandlerChild)
                    signal.signal(signal.SIGTERM, self.SigTermHandlerChild)
                    mirror.Start()
                children.append(result)
            else:
                self.log4py.debug("[ Main (debug) ] Waiting for children to finish")
                signal.pause()
        while len(children) > 0:
            self.log4py.debug("[ Main (debug) ] Waiting for children to finish")
            signal.pause()

        self.DeletePIDFile()
        
        self.log4py.debug("[ Main (debug) ] All mirrors finished")

        if (self.indexfile.directory != None):
            self.log4py.info("[ Main ] Writing Index-file in directory %s" % self.indexfile.directory)
            if self.configfile != None:
                self.ReadConfigFile(self.configfile)
            self.indexfile.Start()



"""

Miscellaneous functions written by Martin Preishuber

"""

from string import strip, lower, zfill, find, atoi, join, replace
from os import getcwd, chdir, mkdir, lstat, unlink, rmdir, utime, wait
from urlparse import urlparse
from time import time
from popen2 import popen3
from re import match

import signal
import os
import string
import copy

# Boolean constants
TRUE = 1
FALSE = 0

# Color constants
BLACK = "30"
RED = "31"
GREEN = "32"
YELLOW = "33"
BLUE = "34"
PURPLE = "35"
AQUA = "36"
WHITE = "37"


"""

String operations

"""

def split(s, sep = None, special = None):
    """ Extended split (doesn't split within special characters). """
    if (special == None):
        if (sep == None):
            return string.split(s)
        else:
            return string.split(s, sep)
    else:
        list = []
        elem = ""
        mode = 0
        for i in range(len(s)):
            if (s[i] == sep):
                if (mode == 0) and (elem != ""):
                    list.append(elem)
                    elem = ""
                elif (mode == 1):
                    elem = "%s%s" % (elem, s[i])
            elif (s[i] == special):
                if (mode == 0):
                    mode = 1
                elif (mode == 1):
                    mode = 0
            else:
                elem = "%s%s" % (elem, s[i])
        list.append(elem)
        return list

def cfill(source, char, width):
    """ Fill a string with a character up to a given width. """
    while len(source) < width:
        source = "%s%s" % (source, char)
    return source

def pfill(source, char, width):
    """ Pad a string with a character up to a given width. """
    while len(source) < width:
        source = char + source
    return source

def lstrip(string):
    """ Strip blanks (NOT whitespaces) on the left side of a string. """
    while (string[0] == " ") and (len(string) > 0):
        string = string[1:]
    return string

def merge(string1, string2):
    """ Merge two strings byte-by-byte. """
    pos1 = 0
    pos2 = 0
    result = ""
    while (len(string1) >= (pos1 + 1)) and (len(string2) >= (pos2 + 1)):
        result = "%s%s%s" % (result, string1[pos1], string2[pos2])
        pos1 = pos1 + 1
        pos2 = pos2 + 1
    if (len(string1) >= (pos1 + 1)):
        result = "%s%s" % (result, string1[pos1:])
    if (len(string2) >= (pos2 + 1)):
        result = "%s%s" % (result, string2[pos2:])
    return result


"""

RE module functions

"""

def unfold(regexp):
    splitted = split(regexp, "\n")
    for i in range(len(splitted)):
        splitted[i] = strip(splitted[i])
    return join(splitted, "")


"""

Type conversion functions

"""

def bool2str(value):
    """ Converts a Boolean value to the corresponding String. """
    if value == FALSE:
        return "False"
    elif value == TRUE:
        return "True"

def str2bool(value):
    """ Converts a String value to the corresponding Boolean. """
    if (lower(value) == "true") or (lower(value) == "yes") or (value == "1"):
        return TRUE
    elif (lower(value) == "false") or (lower(value) == "no") or (value == "0"):
        return FALSE
    else:
        return None

def lng2str(number, prettyprint = TRUE):
    """ Returns a string representation of a number (with a pretty-print option). """
    if (prettyprint == TRUE):
        number = list(str(number))
        if number[-1] == "L":
            number = number[:-1]
        number.reverse()
        dots = divmod(len(number) - 1, 3)[0]
        for i in range(dots):
            number.insert(3 + (i * 3) + i, ".")
        number.reverse()
        return join(number, "")
    else:
        return str(number)

"""

List operations

"""

def striplist(list):
    """ Strip each value of a list. """
    for i in range(len(list)):
        list[i] = strip(list[i])
    return list

def cleanlist(list):
    """ Removes empty entries from a list. """
    newlist = []
    for i in range(len(list)):
        if (list[i] <> ""):
            newlist.append(list[i])
    return newlist

def listdiff(list1, list2):
    """ Get the difference of two lists. """
    difflist = []
    for i in range(len(list1)):
        if (not list1[i] in list2):
            difflist.append(list1[i])
    for i in range(len(list2)):
        if (not list2[i] in list1):
            difflist.append(list2[i])
    return difflist

def listmatch(list, expression):
    """ Returns the matching element number and match object. """
    number = -1
    element = None
    for i in range(len(list)):
        if (match(expression, list[i])):
            number = i
            element = list[i]
            break
    return (number, element)

"""

Time functions

"""

def nicetime(seconds):
    """ Converts time in seconds to a nicely formatted string. """
    minutes, seconds = divmod(seconds, 60)
    return "%s:%s" % (zfill(minutes, 2), zfill(seconds, 2))


"""

File and directory operations

"""

def ls(directory, recursive = FALSE, includepath = FALSE, onlydirectories = FALSE):
    """ Creates a directory listing, optional including full filenames and recursive. """
    longlist = []
    allfiles = os.listdir(directory)
    if (directory == "."): directory = ""
    if ((len(directory) > 0) and (directory[-1] != "/")): directory = "%s/" % directory
    for i in range(len(allfiles)):
        filename = "%s%s" % (directory, allfiles[i])
        longfilename = "%s/%s" % (os.path.abspath(directory), filename)
        if (os.path.isdir(filename)):
            if (includepath == TRUE):
                longlist.append(longfilename)
            else:
                longlist.append(filename)
            if (recursive == TRUE):
                subdirlist = ls("%s/" % filename, recursive, includepath, onlydirectories)
                if (len(subdirlist) > 0):
                    longlist = longlist + subdirlist
        else:
            if (onlydirectories == FALSE):
                if (includepath == TRUE):
                    longlist.append(longfilename)
                else:
                    longlist.append(filename)
    return longlist

def rm(file, maximum = -1, deleted = 0):
    """ Remove a file or directory (recursive) and return the number of directories, files, bytes & links deleted. """
    directories = files = bytes = links = 0L
    if (os.path.islink(file)):
        if (maximum == -1) or ((maximum != -1) and (deleted < maximum)):
            links = links + 1
            unlink(file)
    elif (os.path.isfile(file)):
        filesize = lstat(file)[6]
        if (maximum == -1) or ((maximum != -1) and ((deleted + filesize) < maximum)):
            bytes = bytes + filesize
            files = files + 1
            unlink(file)
    else:
        filelist = os.listdir(file)
        for i in range(len(filelist)):
            result = rm(file + "/" + filelist[i], maximum, bytes)
            directories = directories + result[0]
            files = files + result[1]
            bytes = bytes + result[2]
            links = links + result[3]
        filelist = listdir(file)
        if (maximum == -1) or ((len(filelist) == 0) and (maximum != -1) and (deleted < maximum)):
            rmdir(file)
            directories = directories + 1
    return [directories, files, bytes, links]

def mkdirtree(directory):
    """ Creates a whole direcotry tree. """
    if (directory[-1:] == "/"):
        directory = directory[:-1]
    if (not os.path.exists(directory)):
        pwd = getcwd()
        splitted = split(directory, "/")
        path = "/"
        for i in range(len(splitted)):
            path = "%s/%s" % (path, splitted[i])
            if not os.path.exists(path):
                mkdir(path)
        chdir(pwd)

def du(file):
    """ Calculate the disk usage of a given directory. """
    if (os.path.islink(file)) or (os.path.isfile(file)):
        size = lstat(file)[6]
    else:
        filelist = listdir(file)
        size = 0
        for i in range(len(filelist)):
            size = size + du(file + "/" + filelist[i])
    return size

def which(file):
    """ Check the path to find some specified program. """
    if file == "":
        return ""
    for item in split(os.environ["PATH"], os.pathsep):
        filename="%s%s%s" % (item, os.sep, file)
        if os.path.exists(filename):
            return filename
    return ""

def filecopy(source, target):
    """ Copy a source file to a target file. """
    sourcefile = open(source, "r")
    targetfile = open(target, "w")
    targetfile.write(sourcefile.read())
    targetfile.close()
    sourcefile.close()

def includefile(source, target):
    """ Include a file into a already open file. """
    sourcefile = open(source, "r")
    line = "line"
    while (line != ""):
        line = sourcefile.readline()
        if (line != ""):
            target.write(line)
    sourcefile.close()

def listdir(path, incfiles = TRUE, incdirs = TRUE, inclinks = TRUE):
    """ A modified listdir with some optional parameters. """
    files = os.listdir(path)
    if (incfiles == TRUE) and (incdirs == TRUE) and (inclinks == TRUE):
        cleanfiles = files
    else:
        cleanfiles = []
        if path[-1] != "/":
            path = "%s/" % path
        for i in range(len(files)):
            file = "%s%s" % (path, files[i])
            addFile = TRUE
            if (incfiles == FALSE) and (os.path.isfile(file)):
                addFile = FALSE
            if (incdirs == FALSE) and (os.path.isdir(file)):
                addFile = FALSE
            if (inclinks == FALSE) and (os.path.islink(file)):
                addFile = FALSE
            if (addFile == TRUE):
                cleanfiles.append(files[i])
    cleanfiles.sort()
    return cleanfiles

def touch(filename):
    """ Implements the functionality of the unix command "touch". """
    if (os.path.exists(filename)):
        now = time()
        utime(filename, (now, now))
    else:
        file = open(filename, "w")
        file.close()

def df(directory = ""):
    """ Returns the values of "df" in list form. """
    df_command = which("df")
    df_command = strip("%s --block-size=1 %s" % (df_command, directory))
    output = striplist(cmdoutput(df_command)[1:])
    entry_list = []
    for i in range(len(output)):
        splitted = cleanlist(split(output[i], " "))
        entry = {}
        entry["FILESYSTEM"] = splitted[0]
        entry["SIZE"] = atoi(splitted[1])
        entry["USED"] = atoi(splitted[2])
        entry["AVAILABLE"] = atoi(splitted[3])
        entry["PERCENTAGE"] = atoi(splitted[4][:-1])
        entry["MOUNT_POINT"] = splitted[5]
        entry_list.append(entry)
    return entry_list

def cmpdirs(dict1, dict2):
    """ Compares two dictionaries of lstatfiles and returns a tuple ( {added}, {deleted}, {modified]) """
    added = {} 
    deleted = {}
    modified = {}

    for i in range(len(dict2.keys())):
        key = dict2.keys()[i]
        value = dict2[key]
        if (not key in dict1.keys()):
            added[key] = value
        else:
            if (value[8] != dict1[key][8]):
                modified[key] = value
    for i in range(len(dict1.keys())):
        key = dict1.keys()[i]
        value = dict1[key]
        if (not key in dict2.keys()):
            deleted[key] = value
    return (added, deleted, modified)

def lstatfiles(filelist):
    """ Returns a dictionary {filename: lstat} for a given [filelist]. """
    statdict = {}
    for i in range(len(filelist)):
        statinfo = lstat(filelist[i])
        statdict[filelist[i]] = statinfo
    return statdict

def escapedfilename(filename):
    """ escapes special characters in filenames (currently "`") """
    filename = replace(filename, "`", "\`")
    return filename

"""

Pipe functions

"""

def cmdoutput(command, strip = FALSE, waitpid = TRUE, returnerror = FALSE):
    """ Read the complete output of a command. """
    pipe = popen3(command)

    if (returnerror == TRUE):
        output = pipe[2].readlines()
    else:
        output = pipe[0].readlines()

    if (strip == TRUE):
        output = striplist(output)

    if (pipe[0] != None):
        pipe[0].close()
    if (pipe[1] != None):
        pipe[1].close()
    if (pipe[2] != None):
        pipe[2].close()
    if (waitpid == TRUE):
        os.wait()
    del pipe
    pipe = None
    return output

"""

General functions

"""

def SigChldHandler(signal_number, stack_frame):
    """ Handler, which waits for children to be closed. """

    # Reinstall the sigchild handler, because it gets deleted on libc5/sysv machines
    signal.signal(signal.SIGCHLD, SigChldHandler)
    pid = -1
    while (pid != 0):
        try:
            pid = os.waitpid(0, os.WNOHANG)[0]
        except:
            pid = 0

# ANSI text (fore- & background, bold font)
def ansi(text, fg = WHITE, bg = BLACK, bold = FALSE):
    if (bold == TRUE):
        fg = "%s;1" % fg
    bg = str(atoi(bg) + 10)
    text = "\033[%s;%sm%s\033[0m" % (bg, fg, text)
    return text


"""

Extended version of urlparse (includes username & password)

"""

# extended urlparse function (including username, password)
# returns: <scheme>://<netloc>/<path>;<params>?<query>#<fragment>
# returns: <scheme>://<user>:<password>@<netloc>:<port>/<path>;<params>?<query>#<fragment>

urlparseorg = urlparse
_services_cache = {}

def readservices():
    global _services_cache
    services = {}
    if (len(_services_cache) == 0):
        if (os.path.exists("/etc/services")):
            file = open("/etc/services", "r")
            lines = file.readlines()
            file.close()
            for i in range(len(lines)):
                line = strip(lines[i])
                if (find(line, "#") != -1):
                    line = strip(line[:find(line, "#")])
                if (len(line) > 0):
                    nameportpair = split(line)
                    if (len(nameportpair) > 1):
                        name, port = nameportpair[:2]
                        if find(port, "/") != -1:
                            port, protocol = split(port, "/")
                            services[name] = port
        else:
            services["ftp"] = 21
            services["http"] = 80
        _services_cache = copy.copy(services)
    else:
        services = copy.copy(_services_cache)
    return services

def urlparse(url):
    scheme, netloc, path, params, query, fragment = urlparseorg(url)
    services = readservices()
    user = password = ""
    if services.has_key(scheme):
        port = atoi(services[scheme])
    else:
        port = 0
    if (find(netloc, "@") != -1):
        user, netloc = split(netloc, "@")
    if (find(user, ":") != -1):
        user, password = split(user, ":")
    if (find(netloc, ":") != -1):
        netloc, port = split(netloc, ":")
        port = atoi(port)
    if (path == ""):
        path = "/"
    tuple = scheme, user, password, netloc, port, path, params, query, fragment
    return tuple

"""

Database functions

"""

def db_null_string(s):
    if (s == "") or (s == None):
        return "NULL"
    else:
        return s




"""

Miscellaneous functions to manage unix processes

"""

from commands import getoutput
from string import split, atoi, ljust, strip, find, upper
from tools import cleanlist, TRUE, FALSE
from types import ListType, StringType, DictType

import copy
import sys
import log4py

# Define some necessary constants

FILTERON = 1
FILTEROFF = 0

FILTERMODEEXACT = 1
FILTERMODESUB = 2

# Operating system depended ps parameters

OSTYPEBSD = "BSD"
OSTYPESYSV = "System V"
OSTYPEUNKNOWN = "Unknown"

# The platform can be determined by executing 
#     python -c "import sys;print sys.platform"

osFamily = {}
osFamily[OSTYPEBSD] = ["freebsd4", "bsdos4"]
osFamily[OSTYPESYSV] = ["linux-i386", "linux2"]

osPsParam = {}
osPsParam[OSTYPEBSD] = "-j -p %s"
osPsParam[OSTYPESYSV] = "-f -p %s"
osPsParam[OSTYPEUNKNOWN] = "-f -p %s"

# Definition of the ProcessTable class

class ProcessTable:

    # ProcessTable default settings
    def __init__(self):

        self.processes = {}                                 # List of Dictionaries containing processes
        self.parameter = ""                                 # Parameters passed to read
        self.processmap = {}                                # Dictionary containing the various columns
        self.columnlength = {}                              # Dictionary containing the maximum length of each column
        self.filteredpids = []                              # List of filtered pids
        self.mode = FILTEROFF                               # Current mode
        self.ppidparam = ""
        self.log4py = log4py.Logger().get_instance(self)
        self.log4py.set_loglevel(log4py.LOGLEVEL_DEBUG)

    # Reads the current processes with optional parameters
    def read(self, parameter = "ax"):

        self.processes = {}
        self.parameter = parameter
        self.reread()

    # Re-read processes list (no parameters necessary)
    def reread(self):

        command = "ps %s" % self.parameter
        output = getoutput(command)

        splitted = split(output, "\n")
        header = cleanlist(split(splitted[0], " "))
        processes = splitted[1:]
    
        processentry = {}

        self.processmap = {}
        self.columnlength = {}

        maxheaderindex = 0
        for i in range(len(header)):
            processentry[header[i]] = None
            self.processmap[i] = header[i]
            self.columnlength[i] = len(header[i])

        maxheaderindex = len(header) - 1

        for i in range(len(processes)):
            pid = -1
            splitted = cleanlist(split(processes[i], " "))
            process = copy.copy(processentry)

            for j in range(len(splitted)):

                if (j <= maxheaderindex):
                    if self.processmap[j] == "PID":
                        pid = atoi(splitted[j])
                    process[self.processmap[j]] = splitted[j]
                    if (len(splitted[j]) > self.columnlength[j]):
                        self.columnlength[j] = len(splitted[j])
                else:
                    process[self.processmap[maxheaderindex]] = "%s %s" % (process[self.processmap[maxheaderindex]], splitted[j])
                    
                if (j < maxheaderindex) and (len(splitted[j]) > self.columnlength[j]):
                    self.columnlength[j] = len(splitted[j])

            self.processes[pid] = process

    # Filter for certain conditions in the ProcessTable
    # filters is a dictionary containing parameter:[valuelist]
    def filter(self, filters = None, filtermode = FILTERMODEEXACT):
        self.mode = FILTERON
        self.filteredpids = []

        for i in range(len(self.processes.keys())):

            processid = self.processes.keys()[i]
            process = self.processes[processid]

            processmatches = TRUE
            for j in range(len(filters.keys())):

                filtername = filters.keys()[j]
                filterlist = filters[filtername]
                if type(filterlist) != ListType:
                    filterlist = [filterlist]
                filtername = upper(filtername)

                # special case for different ps implementations
                if (filtername == "CMD") and (process.has_key("COMMAND")):
                    filtername = "COMMAND"

                parameter = process[upper(filtername)]

                for k in range(len(filterlist)):
                    singlefilter = filterlist[k]
                    if (filtermode == FILTERMODEEXACT) or (type(singlefilter) != StringType):
                        if (str(singlefilter) <> str(parameter)):
                            processmatches = FALSE
                    else:
                        if (find(parameter, singlefilter) < 0):
                            processmatches = FALSE

            if (processmatches == TRUE):
                self.filteredpids.append(atoi(process["PID"]))
    
    # Remove applied filter
    def removefilter(self):

        self.filteredpids = []
        self.mode = FILTEROFF

    # Checks, wether a process specified by the given condition(s) exists
    def exists(self, filters = None):

        tempfilter = self.filteredpids
        self.filter(filters)
        exists = (len(self.filteredpids) > 0)
        self.filteredpids = tempfilter

        return exists

    # Returns a dictionary with the specified process
    def process(self, pid):

        if self.processes.has_key(pid):
            return self.processes[pid]
        else:
            return None

    # Print the process table (optional just a list of given pids)
    def printtable(self, list = None):

        if (list == None):
            list = self.processes.keys()
        elif (type(list) == DictType):
            list = list.keys()
        list.sort()

        if len(list) == 0:
            return

        header = ""
        for i in range(len(self.processmap)):
            if i < (len(self.processmap) - 1):
                header = "%s%s " % (header, ljust(self.processmap[i], self.columnlength[i]))
            else:
                header = "%s%s " % (header, self.processmap[i])

        sys.stdout.write("%s\n" % strip(header))

        keys = self.processmap.keys()
        keys.sort()

        for i in range(len(list)):

            if self.processes.has_key(list[i]):
                process = self.processes[list[i]]

                line = ""
                for j in range(len(keys)):

                    key = keys[j]
                    if j < (len(self.processmap) - 1):
                        line = "%s%s " % (line, ljust(process[self.processmap[key]], self.columnlength[j]))
                    else:
                        line = "%s%s " % (line, process[self.processmap[key]])
                sys.stdout.write("%s\n" % strip(line))
            else:
                sys.stdout.write("No process with pid #%s in the internal process table.\n" % list[i])

    # Returns either all pids or all filtered pids
    def pids(self):

        if (self.mode == FILTERON):
            list = self.filteredpids
        else:
            list = self.processes.keys()
        list.sort()
        return list

    # Returns a dictionary of parent pids of given pids, filtered pids or all pids (depends on filter mode)
    def ppids(self, pids = None):
        if (pids == None):
            if (self.mode == FILTERON):
                pids = self.filteredpids
            else:
                pids = self.processes.keys()
        pids.sort()
        ppids = {}
        for i in range(len(pids)):
            pptab = ProcessTable()

            if (self.ppidparam == ""):
                for j in range(len(osPsParam.keys())):
                    opSysFamily = osPsParam.keys()[j]
                    if (osFamily.has_key(opSysFamily)):
                        if (sys.platform in osFamily[opSysFamily]):
                            self.ppidparam = osPsParam[opSysFamily]
                if (self.ppidparam == ""):
                    self.log4py.warn("Unknown operating system platform: %s" % sys.platform)
                    self.ppidparam = osPsParam[OSTYPEUNKNOWN]

            pptab.read(self.ppidparam % pids[i])
            ppid = pptab.process(pptab.pids()[0])["PPID"]
            if not ppids.has_key(ppid):
                ppids[ppid] = [pids[i]]
            else:
                ppids[ppid].append(pids[i])
        return ppids
